-- $Id: archery.lua 2835 2017-04-11 00:13:59Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

local Entity = require 'engine.Entity'

newTalentType {
  name = 'Archery talents',
  type = 'misc/archery',
  description = [[Talents conferred by ranged weapons (bows, slings, crossbows), allowing you to use them]],
}

-- Return true if the player is wielding or carrying suitable ammo.
local function has_ammo(who, t, silent, dry_run)
  local function ammo_cb(o, _, inven)
    return o.type == 'ammo' and o.subtype == t.ammo_type and (inven.id == who.INVEN_INVEN or inven.id == who.INVEN_QUIVER)
  end
  if not util.findInAllInventoriesBy(who, ammo_cb) then
    if not silent then game.logPlayer(who, 'You are out of ammo.') end
    return false
  end
  return true
end

local function fire_speed(self, t)
  return 1/math.max(1, 1 + self.extra_shots)
end

-- cf. do_cmd_fire() in src/cmd2.c
local function shoot_range(self, t)
  local l_inven = self:getInven(self.INVEN_LAUNCHER)
  local launcher = l_inven and l_inven[1]
  if not launcher then return 0 end
  local mul = launcher:ammoMult() + self.extra_might
  return 10 + mul*5
end

-- cf. do_cmd_fire() in src/cmd2.c
local function fire_ammo(self, t)
  -- Sanity check:  make sure we have ammo.
  if not has_ammo(self, t, true, true) then return nil end

  local inv_id, quiver = self.INVEN_QUIVER, self:getInven('QUIVER')
  local stack, item = quiver and #quiver > 0 and quiver[1], 1
  if not stack or stack.subtype ~= t.ammo_type then
    -- Quiver is empty or wrong type; pick something from inventory.
    local function filter_cb(o)
      return o.type == 'ammo' and o.subtype == t.ammo_type
    end
    local function action_cb(o, item)
      self:talentDialogReturn(o, item)
      return false
    end
    local ShowInventory = require 'mod.dialogs.ShowInventory'
    stack, item = self:talentDialog(ShowInventory.new(t.name, self:getInven('INVEN'), filter_cb, action_cb, self))
    -- Allow the player to cancel by not selecting ammo.
    if not stack then return nil end
    -- TODO wield selected ammo?
    inv_id = self.INVEN_INVEN
  end

  local tg = { type='bolt', range=self:getTalentRange(t), talent=t }
  if self:has('PIERCING_SHOTS') then tg.type = 'beam' end
  local tx, ty = self:getTarget(tg)
  if not tx or not ty then return nil end
  local _, _, _, sx, sy = self:canProject(tg, tx, ty)

  local ammo, last = self:getOneObject(inv_id, item, false, {vocal_left=true})
  local p = { 0, ammo=ammo }
  if self:has('PIERCING_SHOTS') then
    -- Piercing shots.
    p.n_pierce = math.max(0, math.floor(self:getSkillLevel('COMBAT')/10) - 1)
  end

  -- Actually fire the ammo.
  self:project(tg, tx, ty, DamageType.ARCHERY, p)

  if p.ammo then
    -- Ammo didn't hit a monster or pierced through every monster it hit.
    -- Either explode or drop to floor at the terminus point.
    if p.ammo.explode_dam_type then
      self:ammoExplode(p.ammo, sx, sy)
    else
      self:ammoDropFloor(ammo, sx, sy)
    end
  end

  if inv_id == self.INVEN_INVEN and not last and config.settings.tome2.auto_quiver then
    -- If there's any ammo left in this stack, user setting says to
    -- auto-wield it.
    self:doWear(inv_id, false, item, stack)
  end

  return true
end

newTalent {
  name = [[Fire Arrow]],
  short_name = 'FIRE_ARROW',
  display_entity = Entity.new { image='talents/fire_arrow.png' },
  type = 'misc/archery',
  mode = 'activated',
  range = shoot_range,
  ammo_type = 'arrows',
  use_speed = fire_speed,
  on_pre_use = has_ammo,
  action = fire_ammo,
  info = [[Fire an arrow using your wielded bow.  If you do not have suitable ammo in your quiver, you will be prompted to choose suitable ammo from your inventory.]],
}

newTalent {
  name = [[Fire Sling Shot]],
  short_name = 'FIRE_SHOT',
  display_entity = Entity.new { image='talents/fire_sling_shot.png' },
  type = 'misc/archery',
  mode = 'activated',
  range = shoot_range,
  ammo_type = 'shots',
  use_speed = fire_speed,
  on_pre_use = has_ammo,
  action = fire_ammo,
  info = [[Fire a sling shot using your wielded sling.  If you do not have suitable ammo in your quiver, you will be prompted to choose suitable ammo from your inventory.]],
}

newTalent {
  name = [[Fire Crossbow Bolt]],
  short_name = 'FIRE_BOLT',
  display_entity = Entity.new { image='talents/fire_crossbow_bolt.png' },
  type = 'misc/archery',
  mode = 'activated',
  range = shoot_range,
  ammo_type = 'bolts',
  use_speed = fire_speed,
  on_pre_use = has_ammo,
  action = fire_ammo,
  info = [[Fire a crossbow bolt using your wielded crossbow.  If you do not have suitable ammo in your quiver, you will be prompted to choose suitable ammo from your inventory.]],
}

newTalent {
  name = [[Throw Boomerang]],
  short_name = 'THROW_BOOMERANG',
  display_entity = Entity.new { image='talents/throw_boomerang.png' },
  type = 'misc/archery',
  mode = 'activated',
  action = function(self, t)
    game.log('Sorry, not implemented (THROW_BOOMERANG)')
  end,
  info = [[Watch this space]],
}
