-- $Id: symbiosis.lua 3522 2018-12-15 20:41:49Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

local Entity = require 'engine.Entity'

newTalentType {
  name = 'Symbiosis powers',
  type = 'skill/symbiosis',
  description = 'Talents conferred by the Symbiosis skill',
}

local function require_symbiote(who, t, silent, dry_run)
  local syms = who:getInven('SYMBIOTE')
  if not syms or #syms == 0 then
    if not silent then game.logPlayer(who, 'You are not in symbiosis.') end
    return false
  end
  return true
end

newTalent {
  name = [[Hypnotize]],
  short_name = 'SYM_HYPNO',
  display_entity = Entity.new { image='talents/hypnotize.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 1,
  mana = 1,
  base_fail = 0,
  stat = 'CON',
  points = 0,
  action = function(self, t)
    local tg = { type='bolt', nowarning=true, range=1, talent=t }
    local tx, ty = self:getTarget(tg)
    if not tx or not ty then return nil end
    local a = game.level.map(tx, ty, Map.ACTOR)
    if not a then
      game.log('No monster at that location.')
      return false
    end
    if a == game.player then
      game.log("That... really wouldn't work.")
      return false
    end
    local name = a:getName{ def_art=true, check_seen=true }
    if not a:has('NEVER_MOVE') then
      game.log('%s is not a never-moving creature.', name:capitalize())
      return false
    end
    if a.faction ~= 'pets' then
      game.log('%s is not your pet.', name:capitalize())
      return false
    end
    if a:has('SPECIAL_GENE') then
      game.log('%s seems immune.', name:capitalize())
      return false
    end
    game.level:removeEntity(a)
    local Object = require 'mod.class.Object'
    Object.resolve_flags.force_mon = a
    local o = game.zone:makeEntityByName(game.level, 'object', 'SYMBIOTE')
    o:identify('full', self)
    o:setFoundHere{ type='custom_with_level', custom='You hypnotized %s' }
    game.level.map:addObject(tx, ty, o)
  end,
  info = function(self, t)
    return [[Places a friendly never-moving creature into a semi-quiescent state, so that it can be merged with your body by 'wearing' it.]]
  end,
}

newTalent {
  name = [[Release Symbiote]],
  short_name = 'SYM_RELEASE',
  display_entity = Entity.new { image='talents/release_symbiote.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 1,
  mana = 1,
  base_fail = 0,
  stat = 'CON',
  points = 0,
  use_item = {
    name = 'Release symbiote',
    close_before = true,
    filter = function(o)
      return o.type == 'misc' and o.subtype == 'symbiote'
    end,
    action = function(self, t, _, inven, item, floor)
      local tg = { type='bolt', nowarning=true, range=5, nolock=true, talent=t }
      local tx, ty = self:getTarget(tg)
      if not tx or not ty then return nil end
      local _, _, _, px, py = self:canProject(tg, tx, ty)
      local x, y = util.findFreeGrid(px, py, 5, true, {[Map.ACTOR]=true})
      if not x or not y then return nil end
      local oo = self:getOneObject(inven, item, floor)
      if not oo then return nil end
      oo.symbiote.force_no_levelup = true
      game.zone:addEntity(game.level, oo.symbiote, 'actor', x, y)
      oo.symbiote.force_no_levelup = nil
      return true
    end,
  },
  action = function(self, t)
    local function sel_cb(o, inven, floor, item, btn, event)
      self:talentDialogReturn(true, {inven=inven, item=item, floor=floor})
      return true
    end
    local ShowEquipInvenFloor = require('mod.dialogs.ShowEquipInvenFloor')
    local ret, p = self:talentDialog(ShowEquipInvenFloor.new('Select symbiote to release', self, t.use_item.filter, sel_cb))
    if not ret then return nil end
    return t.use_item.action(self, t, nil, p.inven, p.item, p.floor)
  end,
  info = function(self, t)
    return [[Restore a hypnotized never-moving creature to its natural state and release it onto a nearby location.]]
  end,
}

newTalent {
  name = [[Charm Never-Moving]],
  short_name = 'SYM_CHARM',
  display_entity = Entity.new { image='talents/charm_never_moving.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 3,
  mana = 2,
  base_fail = 10,
  stat = 'CON',
  points = 0,
  range = 18,
  action = function(self, t)
    local pow = 3 * self:getSkillLevel('SYMBIOSIS')
    return spell.bolt_spell(self, t, DamageType.CHARM_UNMOVING, pow)
  end,
  info = function(self, t)
    local pow = 3 * self:getSkillLevel('SYMBIOSIS')
    local sfx = spell.spell_skill_desc(self, t)
    local ret = [[Tries to manipulate the reactive systems of a never-moving creature to make it friendly (with charm power %d).  Power will increase%s]]
    return ret:format(pow, sfx)
  end,
}

newTalent {
  name = [[Life Share]],
  short_name = 'SYM_SHARE_LIFE',
  display_entity = Entity.new { image='talents/life_share.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 5,
  mana = 5,
  base_fail = 20,
  stat = 'CON',
  points = 0,
  on_pre_use = require_symbiote,
  action = function(self, t)
    -- We assume require_symbiote() has already been called.
    local sym = self:getInven('SYMBIOTE')[1]
    local pct_plyr = self.life / self.max_life
    local pct_mon = sym.symbiote.life / sym.symbiote.max_life
    -- Set both to the average percent of max life.
    local pct = (pct_plyr + pct_mon)/2
    self.life = self.max_life * pct
    sym.symbiote.life = sym.symbiote.max_life * pct
    self.changed = true
    sym.symbiote.changed = true
    return true
  end,
  info = function(self, t)
    return [[Symbiant and symbiote are intermingled; the stronger can support the weaker by taking on some of its damage.

Moves damage between the symbiant and symbiote such that both are equally damaged as a percent of their max life.]]
  end,
}

newTalent {
  name = [[Heal Symbiote]],
  short_name = 'SYM_HEAL',
  display_entity = Entity.new { image='talents/heal_symbiote.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 15,
  mana = 14,
  base_fail = 25,
  stat = 'CON',
  points = 0,
  on_pre_use = require_symbiote,
  action = function(self, t)
    -- We assume require_symbiote() has already been called.
    local sym = self:getInven('SYMBIOTE')[1]
    local pct = 15 + self:getSkillLevel('SYMBIOSIS', 35)
    local hp = (sym.symbiote.max_life * pct)/100
    local name = sym.symbiote:getName{ def_art=true }
    sym.symbiote:heal(hp)
    game.logPlayer(self, '%s is healed.', name:capitalize())
    return true
  end,
  info = function(self, t)
    local pct = 15 + self:getSkillLevel('SYMBIOSIS', 35)
    local ret = [[Heals the symbiote for %d%% of its max life.]]
    return ret:format(pct)
  end,
}

newTalent {
  name = [[Summon Never-Moving Pet]],
  short_name = 'SYM_SUMMON',
  display_entity = Entity.new { image='talents/summon_never_moving_pet.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 30,
  mana = 35,
  base_fail = 40,
  stat = 'CON',
  points = 0,
  action = function(self, t)
    local function f(e) return e:has('NEVER_MOVE') and not e.unique end
    local mons = spell.summon(self, self.x, self.y, {special=f})
    return mons and #mons > 0
  end,
  info = function(self, t)
    return [[Summons a never-moving pet suitable to the current dungeon level.]]
  end,
}

newTalent {
  name = [[Force Symbiosis]],
  short_name = 'SYM_FORCE',
  display_entity = Entity.new { image='talents/force_symbiosis.png' },
  type = 'skill/symbiosis',
  mode = 'activated',
  is_spell = true,
  skill = 'SYMBIOSIS',
  skill_level = 40,
  mana = 60,
  base_fail = 70,
  stat = 'CON',
  points = 0,
  action = function(self, t)
    local tg = { type='bolt', nowarning=true, range=15, talent=t }
    local tx, ty = self:getTarget(tg)
    local a = game.level.map(tx, ty, Map.ACTOR)
    if not a then
      game.log('No monster at that location.')
      return false
    end
    if a == game.player then
      game.log("That... really wouldn't work.")
      return false
    end
    local name = a:getName{ def_art=true, check_seen=true }
    if not a:has('NEVER_MOVE') then
      game.log('%s is not a never-moving creature.', name:capitalize())
      return false
    end
    local talents = util.body_powers(a, 'SYMBIOSIS', 10, 25)
    if not talents or #talents == 0 then
      game.log('%s has no powers you can use.', name:capitalize())
      return false
    end
    local dummy_src = {
      getName = function(not_used) return a:getName{indef_art=true} end,
      worn_talents = talents,
    }
    local force_def = {
      ignore_energy = true,
      ignore_ressources = true,
      force_no_fail = true,
    }
    local UseTalents = require('mod.dialogs.UseTalents')
    return self:talentDialog(UseTalents.new(self, dummy_src, {force_def=force_def, talent_dialog=true}))
  end,
  info = [[Overwhelms the reactive systems of a nearby never-moving creature (friendly or hostile), compelling it to use one of its powers on your behalf.]],
}
