-- $Id: Actor.lua 4288 2021-09-26 03:39:29Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
require 'engine.Entity'
require "engine.Actor"
require "engine.Autolevel"
require "engine.interface.ActorTemporaryEffects"
require "engine.interface.ActorLife"
require "mod.class.interface.ActorProject"
require "engine.interface.ActorLevel"
require "mod.class.interface.ActorStats"
require "mod.class.interface.ActorCallbacks"
require "mod.class.interface.ActorLearn"
require "mod.class.interface.EntityTweakable"
require "engine.interface.ActorTalents"
require 'engine.interface.ActorInventory'
require "engine.interface.ActorResource"
require "engine.interface.ActorFOV"
require 'engine.interface.ActorQuest'
require "mod.class.interface.Combat"
local DamageType = require 'engine.DamageType'
local Map = require "mod.class.Map"
local Faction = require 'engine.Faction'
local Object = require 'mod.class.Object'

module(..., package.seeall, class.inherit(engine.Actor,
					  engine.interface.ActorTemporaryEffects,
					  engine.interface.ActorLife,
					  mod.class.interface.ActorProject,
					  engine.interface.ActorLevel,
					  mod.class.interface.ActorStats,
					  mod.class.interface.ActorCallbacks,
					  mod.class.interface.ActorLearn,
					  engine.interface.ActorTalents,
					  engine.interface.ActorInventory,
					  engine.interface.ActorResource,
					  engine.interface.ActorFOV,
					  engine.interface.ActorQuest,
					  mod.class.interface.EntityTweakable,
					  mod.class.interface.Combat))

-- This is local in engine.interface.ActorResource, so we have to make our
-- own copy.  [cf. :recomputeRegenResources()]
local fast_cache = {}
setmetatable(fast_cache, {__mode='v'})

-- Used by :recomputeRegenResources() below.
_M._no_save_fields.regenResourcesFast = true

-- Used by :canSee() below.
_M._no_save_fields.can_see_cache = true

-- Used by the running/resting infrastructure.
_M._no_save_fields.running = true
_M._no_save_fields.resting = true

-- Used internally by the ActorProject infrastructure, I think.
_M._no_save_fields.__project_source = true

function _M:init(t, no_default)
  self.energyBase = 0

  engine.Actor.init(self, t, no_default)
  engine.interface.ActorTemporaryEffects.init(self, t)
  engine.interface.ActorLife.init(self, t)
  mod.class.interface.ActorProject.init(self, t)
  engine.interface.ActorTalents.init(self, t)
  engine.interface.ActorResource.init(self, t)
  mod.class.interface.ActorStats.init(self, t)
  mod.class.interface.ActorCallbacks.init(self, t)
  mod.class.interface.ActorLearn.init(self, t)
  engine.interface.ActorInventory.init(self, t)
  engine.interface.ActorLevel.init(self, t)
  engine.interface.ActorFOV.init(self, t)

  -- Short-circuit the engine's initial forced level-up mechanism, which
  -- doesn't work quite the way we want.
  self.level = t.level or 1
  self.start_level = self.level

  self:recomputeRegenResources()
  self:resetCanSeeCache()
end

function _M:loaded()
  engine.Actor.loaded(self)
  self:recomputeRegenResources()
end

function _M:resolveSource()
  if self.summoner then
    return self.summoner.resolveSource and self.summoner:resolveSource() or self.summoner
  end
  if self.symbiosis then
    return self.symbiosis.host.resolveSource and self.symbiosis.host:resolveSource() or self.symbiosis.host
  end
  return self
end

function _M:resolve(t, last, on_entity, key_chain)
  -- Dodgy Hack(TM):  resolve() calls itself, and we only want to do our
  -- special stuff once at the top level.
  if self.nested then
    engine.Actor.resolve(self, t, last, on_entity, key_chain)
    return
  end
  self.nested = true

  engine.Actor.resolve(self, t, last, on_entity, key_chain)

  -- This isn't getting handled right, and I can't figure out how to
  -- disentangle it; ham-hand it manually here.
  if last then
    self.life = self.max_life
  end

  self.nested = nil
end

function _M:getName(t)
  t = t or {}

  if t.check_seen and self ~= game.player and game.player.x and game.player.y then
    local los = game.player:hasLOS(self.x, self.y)
    if not game.player:canSee(self, los, (los and 100 or 0)) then
      return 'something'
    end
  end

  local name = self.name
  local no_your = t.no_your or (t.cite_given_name and self.given_name)
  local pfx, sfx = '', ''
  if self.ego then
    local NPCEgo = require 'mod.class.NPCEgo'
    local ego = NPCEgo:get(self.ego)
    if ego.prefix then
      pfx = ego.name .. ' ' .. pfx
    else
      sfx = sfx .. ' ' .. ego.name
    end
  end
  if self.faction == 'pets' and not self.unique and not no_your then
    pfx = 'your ' .. pfx
  elseif t.indef_art and not self.unique and not self.player then
    pfx = (name:match('^[AEIOUaeiou]') and 'an ' or 'a ') .. pfx
    if self.given_name and t.cite_given_name then
      sfx = sfx .. ' named ' .. self.given_name
    end
  elseif t.def_art and not self.unique and not self.player then
    pfx = 'the ' .. pfx
    if self.given_name and t.cite_given_name then
      sfx = sfx .. ' ' .. self.given_name
    end
  end
  if t.plural and not self.unique and not self.player and not self.self_plural then
    if self.plural_name then
      name = self.plural_name
    elseif name:find(' of ') then
      name = name:gsub(' of ', 's of ', 1)
    elseif name:find('y$') then
      name = name:gsub('y$', 'ies', 1)
    elseif name:find('ch$') then
      -- Patriarch is an exception to this, covered by its plural_name
      -- field.
      name = name .. 'es'
    elseif name:find('ss$') then
      name = name .. 'es'
    --[[ These are now covered by the 'self_plural' flag.
    elseif name:find('fish$') then
      -- Already plural.
    elseif name:find(' coins$') then
      -- Already plural, I suppose.
    elseif name:find(' druj$') then
      -- Apparently the plural of 'druj' is 'druj'.  Who knew?
    ]]
    elseif name:find(' vortex$') then
      name = name:gsub(' vortex$', ' vortices', 1)
    else
      name = name .. 's'
    end
  end
  if self.given_name and not t.no_given_name and not t.cite_given_name then
    name = self.given_name
    pfx, sfx = ''
  end
  return pfx .. name .. sfx
end

function _M:setName(newname)
  self.given_name = newname
  self.changed = true
end

function _M:pronoun(p, t)
  t = t or {}
  local male, female = self:has('MALE'), self:has('FEMALE')
  local cant_see = false
  if t.check_seen then
    local los = game.player:hasLOS(self.x, self.y)
    cant_see = not game.player:canSee(self, los, (los and 100 or 0))
  end

  if p == 'subject' then
    return cant_see and 'it' or male and 'he' or female and 'she' or 'it'
  elseif p == 'object' then
    return cant_see and 'it' or male and 'him' or female and 'her' or 'it'
  elseif p == 'possessive' then
    return cant_see and 'its' or male and 'his' or female and 'her' or 'its'
  elseif p == 'reflexive' then
    return cant_see and 'itself' or male and 'himself' or female and 'herself' or 'itself'
  end
end

function _M:actBase()
  self.energyBase = self.energyBase - game.energy_to_act

  -- Regen resources
  self:regenLife()
  self:regenResources()
  -- Compute timed and sustained effects
  self:timedEffects()

  self:on_global_turn()

  self:triggerHook{'Actor:actBase'}
  self:fireCallbacks('on_global_turn')

  -- Cooldown talents after callbacks, as some of them can deactivate
  -- sustains.
  self:cooldownTalents()
end

-- Overridden in Player.
function _M:on_global_turn()
end

function _M:on_pickup_object(o, n)
  -- Clear any hallucination or shimmer visuals.
  o:replaceDisplay('hallucination', nil)
  o:replaceDisplay('shimmer', nil)
  -- Forget its map mark.
  o:markForMap(false)
  self:fireCallbacks('on_pickup_floor', o, n)
end

function _M:act()
  if not engine.Actor.act(self) then return end

  self.changed = true

  self:resetCanSeeCache()

  self:fireCallbacks('on_act')

  -- Still enough energy to act ?
  if self.energy.value < game.energy_to_act then return false end

  return true
end

-- Overriden to use :regenMana() et al. methods if present.
function _M:regenResources()
  for _, def in ipairs(self.resources_def) do
    local method = 'regen'..def.short_name:lower():capitalize()
    if self[method] then
      self[method](self)
    elseif def.regen_prop and self[def.regen_prop] then
      self[def.short_name] =
	util.bound(self[def.short_name] + self[def.regen_prop],
		   self[def.minname], self[def.maxname])
    end
  end
end

function _M:recomputeRegenResources()
  if not self._no_save_fields.regenResourcesFast then return end
  local f = 'return function(self) '

  for _, def in ipairs(self.resources_def) do
    local method = 'regen'..def.short_name:lower():capitalize()
    if self[method] then
      f = f .. 'self:' .. method .. '() '
    elseif def.regen_prop and self[def.regen_prop] then
      f = f .. ('self.%s = util.bound(self.%s + self.%s, self.%s, self.%s) '):
	       format(def.short_name, def.short_name, def.regen_prop,
		      def.minname, def.maxname)
    end
  end
  f = f .. ' end'
  if not fast_cache[f] then fast_cache[f] = loadstring(f)() end
  self.regenResourcesFast = fast_cache[f]
end

-- Wrapper function to add a flyer to an actor.
function _M:addFlyer(duration, xvel, yvel, str, color, bigfont)
  local sx, sy = game.level.map:getTileToScreen(self.x, self.y)
  game.flyers:add(sx, sy, duration, xvel, yvel, str, color, bigfont)
end

--- Attach or remove a display callback
-- Defines particles to display
function _M:defineDisplayCallback()
  if not self._mo then return end

  -- Sneaky Hack(TM):  The displayCallback()s we're about to hang on
  -- self._mo and possibly self._last_mo wil have backreferences to self,
  -- creating a reference loop that would prevent GC.  We break the loop by
  -- passing in self via a weak reference that GC can break.
  -- [tip o' the hat to DarkGod, who uses this trick in T4]
  local weak = setmetatable({[1]=self}, {__mode='v'})

  local ps = self:getParticlesList()

  local f_self = nil
  local f_danger = nil
  local f_friend = nil
  local f_enemy = nil
  local f_neutral = nil

  local function particles(x, y, w, h)
    local self = weak[1]
    if not self then return end

    local e
    for i = 1, #ps do
      e = ps[i]
      e:checkDisplay()
      if e.ps:isAlive() then e.ps:toScreen(x + w / 2, y + h / 2, true, w / game.level.map.tile_w)
      else self:removeParticles(e)
      end
    end
  end

  local function tactical(x, y, w, h)
    local self = weak[1]
    if not self then return end

    -- Tactical info
    if game.level and game.level.map.view_faction then
      local map = game.level.map

      if not f_self then
	f_self = map.tilesTactic:get(nil, 0,0,0, 0,0,0, map.faction_self)
	f_danger = map.tilesTactic:get(nil, 0,0,0, 0,0,0, map.faction_danger)
	f_friend = map.tilesTactic:get(nil, 0,0,0, 0,0,0, map.faction_friend)
	f_enemy = map.tilesTactic:get(nil, 0,0,0, 0,0,0, map.faction_enemy)
	f_neutral = map.tilesTactic:get(nil, 0,0,0, 0,0,0, map.faction_neutral)
      end

      if self.faction then
	local friend
	if not map.actor_player then friend = Faction:factionReaction(map.view_faction, self.faction)
	else friend = map.actor_player:reactionToward(self) end

	local f_mark = config.settings.tome2.faction_markers
	local h_bar = config.settings.tome2.health_bars
	local cat, f_tac

	-- Determine which category the actor is in, including which
	-- faction marker to use and which side to draw health bars on.
	if self == map.actor_player then
	  f_tac, cat = f_self, 'self'
	elseif friend > 0 then
	  f_tac, cat = f_friend, 'friend'
	elseif friend == 0 then
	  f_tac, cat = f_neutral, 'neutral'
	elseif map:faction_danger_check(self) then
	  f_tac, cat = f_danger, 'danger'
	else
	  f_tac, cat = f_enemy, 'enemy'
	end

	if f_tac and cat and f_mark[cat] then
	  f_tac:toScreen(x, y, w, h)
	end
	if cat and h_bar[cat] and type(self.life) == 'number' then
	  local lp = math.max(0, self.life) / self.max_life + 0.0001
	  -- Health bar is on the bottom of the tile for now, with faction
	  -- markers on the left or right.  We may want to find a way to
	  -- let the tileset control this.
	  local dw, dh = w/16, h/16
	  local lx, ly, lw, lh = x + dw, y + h - dh, w - 2*dw, dh
	  local lpx, lpy, lpw, lph = x + dw, y + h - dh, (w - 2*dw)*lp, dh
	  local rb, gb, bb, rf, gf, bf
	  local ab, af = 128, 255
	  if lp > 0.75 then -- green
	    rb, gb, bb, rf, gf, bf = 128, 189, 57, 50, 220, 77
	  elseif lp > 0.5 then -- yellow
	    rb, gb, bb, rf, gf, bf = 175, 175, 10, 240, 252, 35
	  elseif lp > 0.25 then -- orange
	    rb, gb, bb, rf, gf, bf = 185, 88, 0, 255, 156, 21
	  else -- red
	    rb, gb, bb, rf, gf, bf = 167, 55, 39, 235, 0, 0
	  end
	  core.display.drawQuad(lx, ly, lw, lh, rb, gb, bb, ab)
	  core.display.drawQuad(lpx, lpy, lpw, lph, rf, gf, bf, af)
	end
      end
    end
  end

  if self._mo == self._last_mo or not self._last_mo then
    self._mo:displayCallback(function(x, y, w, h)
      tactical(x, y, w, h)
      particles(x, y, w, h)
      return true
    end)
  else
    self._mo:displayCallback(function(x, y, w, h)
      tactical(x, y, w, h)
      return true
    end)
    self._last_mo:displayCallback(function(x, y, w, h)
      particles(x, y, w, h)
      return true
    end)
  end
end

function _M:move(x, y, force)
  local moved = false
  local ox, oy = self.x, self.y
  -- Clean this up aggressively.
  local moving_dir = self.moving_dir
  self.moving_dir = nil

  if force or self:enoughEnergy() then
    -- Check for confusion or random movement flag.
    local rand_prob = self.ai_state and self.ai_state.random_move or 0
    local rand_move = self:hasEffect(self.EFF_CONFUSED) or rng.range(1,100) <= rand_prob
    -- Slight Hack(TM):  If targeting player and can't see target, move
    -- randomly.
    -- [cf. process_monster() in src/melee2.c]
    local target = self.ai_target and self.ai_target.actor
    if target and target == game.player then
      rand_move = rand_move or not self:canSee(target)
    end
    if rand_move and self.x and self.y then
      x, y = self.x + rng.range(-1,1), self.y + rng.range(-1,1)
    end

    local no_move_msg =
      self:has('NEVER_MOVE') and 'You cannot move.' or
      self:has('IMMOVABLE') and 'You cannot move.' or
      self:hasEffect(self.EFF_TREE_ROOTS) and 'Your tree roots prevent you from moving.'
    if not force and no_move_msg then
      -- The collision code will detect an attempted attack for us.
      if not game.level.map:checkAllEntities(x, y, 'block_move', self, true) then
	if moving_dir and self:has('IMMOVABLE') and self.descriptor and self.descriptor.race == 'DeathMold' then
	  -- Slight Hack(TM):  DeathMold movement
	  spell.teleport_directed(self, 10, self.move_dir)
	else
	  game.logPlayer(self, no_move_msg)
	end
      end
    elseif not force and self:has('PROBABILITY_TRAVEL') and math.abs(x-ox) <= 1 and math.abs(y-oy) <= 1 and game.level.map:checkEntity(x, y, Map.TERRAIN, 'wall') then
      moved = spell.passwall(self, x - ox, y - oy, true, true)
    else
      moved = engine.Actor.move(self, x, y, force)
    end

    if not force and moved and (self.x ~= ox or self.y ~= oy) then
      if self.pushed_past then
	-- If we pushed past a monster, put it where we were.
	local mon = self.pushed_past
	self.pushed_past = nil
	if mon.x == self.x and mon.y == self.y then
	  print(('[MOVE] %s from %d,%d pushes past %s @%d,%d'):format(self:getName(), ox, oy, mon:getName(), mon.x, mon.y))
	  mon:move(ox, oy, true)
	  self:learnFlags('flags', 'MOVE_BODY')
	  game.logPlayer(self, 'You push past %s.', mon:getName{def_art=true, check_seen=true})
	end
      end

      if self.squished then
	-- If we squished a monster, remove it.
	local mon = self.squished
	self.squished = nil
	if mon.x == self.x and mon.y == self.y then
	  print(('[MOVE] %s from %d,%d squishes %s @%d,%d'):format(self:getName(), ox, oy, mon:getName(), mon.x, mon.y))
	  game.level.map:remove(mon.x, mon.y, Map.ACTOR, mon)
	  if game:hasEntity(mon) then game:removeEntity(mon) end
	  self:learnFlags('flags', 'KILL_BODY')
	end
      end

      if not self.did_energy then self:useEnergy(self:energyToMove()) end

      if game.zone.worldmap then
	-- If we've moved in the world map, adjust our wilderness entry
	-- point accordingly.  Note that wilderness and town zones are
	-- hardcoded to 198x66 size, and the 0-indexed borders (x=0 or 197,
	-- y=0 or 65) are special "crossing" zones.
	self.to_wild_xy = {
	  x = (self.x < ox and 196 or self.x > ox and 1 or 99),
	  y = (self.y < oy and  64 or self.y > oy and 1 or 33),
	}
      end
    end
  end

  self:fireCallbacks('on_move', moved, force, x, y, ox, oy)
  self:triggerHook{'Actor:move', moved=moved, force=force, x=x, y=y, ox=ox, oy=oy}

  self.did_energy = nil

  return moved
end

-- Energy required to move.  This is overrided in Player to handle the
-- worldmap.
function _M:energyToMove()
  return game.energy_to_act
end

function _M:block_move(x, y, e, can_attack)
  if e and e ~= self and can_attack then
    -- If it can push past us, let it do so, and make a note that it did so.
    if e.canPushPast and e:canPushPast(self) then
      e.pushed_past = self
      return false
    end
    -- Similarly if it can squish us.
    if e.canSquish and e:canSquish(self) then
      e.squished = self
      return false
    end
    -- Otherwise, try attacking it.
    e:attack(self)
    return 'attack'
  end
  return true
end

-- Returns true if this Actor can push past the provided Actor by moving
-- into it.
-- Overridden in Player and NPC
function _M:canPushPast(actor)
  return false
end

-- Return true if this Actor can destroy the provided Actor by moving into it.
-- Overridden in NPC (poor player don't get to squish nobody... *pout*)
function _M:canSquish(actor)
  return false
end

-- Override the engine's Actor:teleportRandom() to match T2's conditions
-- for a suitable destination grid.  [cf. teleport_away() in src/spells1.c]
function _M:teleportRandom(x, y, dist, min_dist)
  local cands = {}
  dist = math.floor(dist)
  min_dist = math.floor(min_dist or 0)

  for i = x - dist, x + dist do
    for j = y - dist, y + dist do
      -- Target grid must:
      --   - be in map bounds
      --   - be within the specified min/max distance
      --   - be a floor with no actors on it [cf. Grid:empty_bold()]
      --   - not have a glyph of warding on it
      --   - not have the 'no_teleport' or 'vault' attribute.
      --   - be a legal grid for the actor to enter [not present in the T2
      --     code, but makes sense]
      if game.level.map:isBound(i, j) and
	 core.fov.distance(x, y, i, j) <= dist and
	 core.fov.distance(x, y, i, j) >= min_dist and
	 self:canMove(i, j) and
	 not game.level.map:checkEntity(i, j, Map.TERRAIN, 'no_teleport') and
	 game.level.map:checkEntity(i, j, Map.TERRAIN, 'empty_bold') and
	 not game.level.map.attrs(i, j, "no_teleport") and
	 not game.level.map.attrs(i, j, "vault")
      then
	 cands[#cands+1] = {i,j}
      end
    end
  end

  -- TODO:  The T2 code tries several times, expanding dist and reducing
  -- min_dist until it finds a suitable location.
  if #cands == 0 then return false end
  local pos = cands[rng.range(1, #cands)]
  return self:move(pos[1], pos[2], true)
end

-- Overridden in Player and NPC.
function _M:getDesc(t)
  return '[there should be something here...]'
end

function _M:tooltip(x, y, seen_by)
  -- Don't show a tooltip for an Actor we can't see.
  if seen_by and not seen_by:canSee(self) then return end
  return self:getDesc{with_level=true}
end

-- Gain experience.
-- We override this from engine.interface.ActorLevel to prevent it from
-- subtracting out the exp from the previous level.
function _M:gainExp(value, regain)
  self.changed = true
  self.exp = math.max(0, self.exp + value)
  local leveled_up = false
  if value > 0 then
    local next_lev = self:getExpChart(self.level + 1)
    while next_lev and self.exp >= next_lev and
	  (not self.actors_max_level or self.level < self.actors_max_level) and
	  (not self.max_level or self.level < self.max_level)
    do
      self.level = self.level + 1
      local newlev = self:levelup()
      leveled_up = newlev or leveled_up
      next_lev = self:getExpChart(self.level + 1)
      self:triggerHook{ 'Actor:levelup', actor=self, newlev=newlev }
    end
  elseif value < 0 then
    local cur_lev = self:getExpChart(self.level)
    while self.level > 1 and cur_lev and self.exp < cur_lev do
      self.level = self.level - 1
      self:levelup(true)
      cur_lev = self:getExpChart(self.level)
    end
  end
  if leveled_up then
    self:levelupPost()
    self:triggerHook{ 'Actor:levelupPost', actor=self }
  end
end

-- Stub.
function _M:levelupPost()
end

-- Forces an actor to levelup to "lev"
-- As above, we override this from engine.interface.ActorLevel to prevent
-- it from subtracting out the exp from the previous level.
function _M:forceLevelup(lev)
  -- Vaguely Dodgy Hack(TM):  Sometimes this gets called when we don't want
  -- it (notably actor level adjustments in zones); this flag tells us to
  -- do nothing in that case.
  if self.force_no_levelup then return end

  if self.level < lev then
    -- Since we're not subracting out exp from the previous level, we don't
    -- have to do this a level at a time; we can just "gain" enough exp to
    -- get us to the desired level in one proverbial swell foop. ;)
    -- This way all the usual hooks will get triggered.
    self:gainExp(self:getExpChart(lev) - self.exp)
  end
end

-- Mostly duplicated from engine.interface.ActorLife:takeHit(); we change
-- it to call our Actor:getName() instead of using the raw entity name.
function _M:takeHit(value, src, death_note)
  if self.onTakeHit then value = self:onTakeHit(value, src) end
  self.life = self.life - value
  self.changed = true
  if self.life <= self.die_at and not self.dead then
    if src.on_kill and src:on_kill(self) then return false, value end
    local src_name = src.getName and src:getName{def_art=true, check_seen=true} or src.name or 'something'
    local tgt_name = self:getName{def_art=true, check_seen=true}
    game.logSeen(self, "#{bold}#%s killed %s!#{normal}#", src_name:capitalize(), tgt_name)
    return self:die(src, death_note), value
  end

  if self.onTakeHitPost then self:onTakeHitPost(value, src) end

  return false, value
end

function _M:onTakeHit(value, src)
  -- Quest givers and rescuees are invulnerable.
  if self:has('NO_DEATH') then return 0 end

  local p = { damage=value }
  -- Dodgy and Ugly Hack(TM):  Use the weapon value passed in from the
  -- damage projector, if any.
  self:fireCallbacks('on_take_hit_pre', src, p, self.__hack_melee_weapon)
  value = p.damage

  -- Sneaky Hack(TM):  Quest monsters and non-friendly uniques can only be
  -- killed by the player.
  if src ~= game.player and (self.quest_mon or (self.unique and self:reactionToward(game.player) <= 0)) then
    value = math.min(value, self.life - self.die_at - 1)
  end

  return value
end

function _M:onTakeHitPost(value, src)
  self:fireCallbacks('on_take_hit_post', src, value)
end

-- Juggle temporary effect callbacks.
function _M:on_temporary_effect_added(eff_id, e, p)
  self:installCallbacks(e, 'callbacks', eff_id, 'effect')
end

function _M:on_temporary_effect_removed(eff_id, e, p)
  self:uninstallCallbacks(e, 'callbacks', eff_id)
end

function _M:onAddObject(o)
  engine.interface.ActorInventory.onAddObject(self, o)
  self:installCallbacks(o, 'carried_callbacks', o, 'object')
end

function _M:onRemoveObject(o)
  engine.interface.ActorInventory.onRemoveObject(self, o)
  self:uninstallCallbacks(o, 'carried_callbacks', o)
end

function _M:onWear(o, inven_id)
  engine.interface.ActorInventory.onWear(self, o, inven_id)
  self:installCallbacks(o, 'worn_callbacks', o, 'object')
end

function _M:onTakeoff(o, inven_id)
  engine.interface.ActorInventory.onTakeoff(self, o, inven_id)
  self:uninstallCallbacks(o, 'worn_callbacks', o)
end

-- "Superload" engine.interface.ActorInventory.takeoffObject() to add a log
-- message as appropriate.
function _M:takeoffObject(inven_id, item, vocal)
  local o = engine.interface.ActorInventory.takeoffObject(self, inven_id, item)
  if o and vocal then
    -- We don't have to pass check_seen to Actor:getName() here, since that
    -- would be redundant with game.logSeen().
    game.logSeen(self, '%s takes off: %s.', self:getName():capitalize(), o:getName{do_color=true})
  end
  return o
end

-- Returns false if :addObject() would fail for lack of inventory room.
function _M:canAddObject(inven_id, o, no_unstack)
  local inven = self:getInven(inven_id)
  local stackable, stack_limit = o and o:stackable(), self:invenStackLimit(inven_id)

  -- Easy if we have room.
  if #inven < inven.max then return true end
  -- No?  Look for something we can stack with.
  if stackable and not no_unstack then
    for i, io in ipairs(inven) do
      if o:canStack(io) and io:getNumber() + o:getNumber() < stack_limit then
	return true
      end
    end
  end
  -- I got nothin'...
  return false
end

-- Add an object to an inventory, or drop it on the floor if there is no
-- room.  Passes through the returned values from :addObject().
function _M:addOrDropObject(inven_id, o, no_unstack, force_item, t)
  t = t or {}
  local added, slot, rs = self:addObject(inven_id, o, no_unstack, force_item)
  if not added then
    -- We bypass some of the drop-related callbacks since this was never in
    -- an inventory.
    local x, y, long = util.drop_loc_near(self.x, self.y, 10, o.unique)
    if not x or not y then
      if t.vocal_drop then
	game.logSeen(self, "%s disappears!", o:getName{do_color=true})
      end
      return o
    end
    if long and t.vocal_drop then
      -- Special mention if we had to try hard to find a drop spot.
      game.logSeen(self, "%s vanishes!", o:getName{do_color=true})
    end

    local ok, idx = game.level.map:addObject(x, y, o)

    if t.vocal_drop then game.logSeen(self, "%s drops on the floor: %s.", self.name:capitalize(), o:getName{do_color=true}) end
    if ok and game.level.map.attrs(self.x, self.y, "on_drop") then
      game.level.map.attrs(self.x, self.y, "on_drop")(self, self.x, self.y, idx, o)
    end
  else
    local inven = self:getInven(inven_id)
    if t.vocal_add and inven and slot and inven[slot] then
      game.logPlayer(self, 'You have %s.', inven[slot]:getName{do_color=true})
    end
  end
  return added, slot, rs
end

-- Override engine.interface.ActorInventory:dropFloor() to use
-- util.drop_loc_near().
function _M:dropFloor(inven, item, vocal, all)
  local o = self:getInven(inven)[item]
  if not o then
    if vocal then game.logSeen(self, "There is nothing to drop.") end
    return
  end
  if o:check("on_drop", self) then return false end

  o = self:removeObject(inven, item, all)

  self:onDropObject(o)

  local x, y, long = util.drop_loc_near(self.x, self.y, 10, o.unique)
  if not x or not y then
    if vocal then
      game.logSeen(self, "%s disappears!", o:getName{do_color=true})
    end
    return o
  end
  if long and vocal then
    -- Special mention if we had to try hard to find a drop spot.
    game.logSeen(self, "%s vanishes!", o:getName{do_color=true})
  end

  local ok, idx = game.level.map:addObject(self.x, self.y, o)

  if vocal then game.logSeen(self, "%s drops on the floor: %s.", self.name:capitalize(), o:getName{do_color=true}) end
  if ok and game.level.map.attrs(self.x, self.y, "on_drop") then
    game.level.map.attrs(self.x, self.y, "on_drop")(self, self.x, self.y, idx, o)
  end
  return o
end

function _M:learnTalent(t_id, force, nb)
  local knew_talent = self:knowTalent(t_id)
  local ok, err = engine.interface.ActorTalents.learnTalent(self, t_id, force, nb)
  if not ok then return ok, err end

  if not knew_talent then
    local t = self:getTalentFromId(t_id)
    self:installCallbacks(t, 'callbacks', t_id, 'talent')
  end
  return ok, err
end

function _M:unlearnTalent(t_id, nb)
  local knew_talent = self:knowTalent(t_id)
  local ok, err = engine.interface.ActorTalents.unlearnTalent(self, t_id, nb)
  if not ok then return ok, err end

  if knew_talent then
    local t = self:getTalentFromId(t_id)
    self:uninstallCallbacks(t, 'callbacks', t_id, 'talent')
  end
  return ok, err
end

function _M:die(src, death_note)
  -- See if any of our callbacks let us not die.
  local p = { no_die=false }
  if self.on_die_pre and self:on_die_pre(src) then p.no_die = true end
  self:fireCallbacks('on_die_pre', src, p)
  if p.no_die then
    -- We get to live; patch us back up. [cf. src/dungeon.c lines 5947-6011]
    if self.death_count then self.death_count = self.death_count + 1 end
    self.life = self.max_life
    if self.sanity then self.sanity = self.max_sanity end
    if self.mana then self.mana = self.max_mana end
    self:removeEffect(self.EFF_BLIND)
    self:removeEffect(self.EFF_CONFUSED)
    self:removeEffect(self.EFF_POISON)
    self:removeEffect(self.EFF_AFRAID)
    self:removeEffect(self.EFF_PARALYZED)
    self:removeEffect(self.EFF_HALLUCINATION)
    self:removeEffect(self.EFF_STUN)
    self:removeEffect(self.EFF_CUT)
    self:removeEffect(self.EFF_BLACK_BREATH)
    -- TODO don't kick Necromancers into undead form.
    if self.setFood then self:setFood(util.FOOD.max - 1) end
    self:removeEffect(self.EFF_RECALL)
    if self.recalcEverything then self:recalcEverything() end
    return false
  end

  local could_see = not self.player and (game.player:hasLOS(self.x, self.y) and game.player:canSee(self)) or (self.symbiosis and self.symbiosis.host.player)

  engine.interface.ActorLife.die(self, src, death_note)
  local rsrc = src.resolveSource and src:resolveSource()

  -- Gives the killer some exp for the kill
  if rsrc and rsrc.gainExp then
    rsrc:gainExp(self:worthExp(rsrc))
  end

  -- We fire 'on_kill' callbacks separately for the proximate and
  -- "resolved" killers.
  if rsrc and rsrc.fireCallbacks then
    rsrc:fireCallbacks('on_kill', self, nil)
  end
  if src.fireCallbacks and (not rsrc or src ~= rsrc) then
    src:fireCallbacks('on_kill', self, rsrc)
  end

  -- Find a drop location for our stuff.
  local dropx, dropy = util.drop_loc_near(self.x, self.y, 3)
  if not dropx then
    -- TODO I guess stuff should disappear?
    dropx, dropy = self.x, self.y
  end

  local any_drop, n_drop_gold, n_drop_item = false, 0, 0
  local invens = {}
  for id, inven in pairs(self.inven) do
    invens[#invens+1] = inven
  end
  -- [Not sure why we're sorting these; I'm following T4's Actor:die()]
  local f = function(a, b) return a.id ~= 1 and (b.id == 1 or a.id < b.id) end
  table.sort(invens, f)
  for id, inven in pairs(invens) do
    for i = #inven, 1, -1 do
      local o = inven[i]
      self:removeObject(inven, i, true)
      game.level.map:addObject(dropx, dropy, o)
      any_drop = true
      if o.type == 'money' then
	n_drop_gold = n_drop_gold + 1
      else
	n_drop_item = n_drop_item + 1
      end
    end
  end
  -- Special drops for some monsters.
  local o_spec = self.drop_special and self:drop_special()
  if o_spec then
    o_spec:setFoundHere {
      type = 'mon_drop',
      mon_name = self:getName{ indef_art=true },
    }
    game.level.map:addObject(dropx, dropy, o_spec)
    any_drop = true
  end
  if self.drop_corpse and rng.percent(10 + game.player:getSkillLevel('CORPSE_PRESERVATION', 75)) then
    -- Drop corpse/skeleton.
    local what = self.drop_corpse.corpse and 'CORPSE' or self.drop_corpse.skeleton and 'SKELETON'
    if what then
      Object.resolve_flags = { force_mon=self }
      local o_c = game.zone:makeEntityByName(game.level, 'object', what, true)
      if o_c then
	o_c:setFoundHere {
	  type = 'mon_drop',
	  mon_name = self:getName{ indef_art=true },
	}
	game.level.map:addObject(dropx, dropy, o_c)
	any_drop = true
      end
    end
  end
  if (could_see or self.unique) and (n_drop_gold > 0 or n_drop_item > 0) then
    self:learnFlags(true, 'drops', n_drop_item, n_drop_gold)
  end
  self.inven = {}
  if not self.player and any_drop and dropx == game.player.x and dropy == game.player.y then
    game.log('You feel something roll beneath your feet.')
  end
  if any_drop and game.level.map.attrs(dropx, dropy, 'obj_seen') then
    -- New objects on this spot that the player hasn't seen.
    game.level.map.attrs(dropx, dropy, 'obj_seen', false)
  end

  if (src and src.player) or (rsrc and rsrc.player) then
    self:recordKill()
    world:gainAchievement('KILL_1000', game.player, self)
    world:gainAchievement('KILL_5000', game.player, self)
    world:gainAchievement('KILL_10000', game.player, self)
  end

  -- Explode!
  local AttackType = require 'mod.class.AttackType'
  for _, atk in ipairs(self.attacks or {}) do
    if atk.method == AttackType.EXPLODE then
      local dam = {
	(atk.dam and rng.dice(atk.dam[1], atk.dam[2]) or 0),
	grid = true,
	object = true,
	display_src = { name=self:convertTalentMessage(src, "@source@'s explosion") },
      }
      game.logSeen(self, self:convertTalentMessage(src, 'WSource@ explodes!'))
      spell.ball_spell(self, nil, atk.type, dam, 3, self.x, self.y, 0, spell.simple_blast_particle)
      break
    end
  end

  if self.quest_on_die then
    local q = game.player:hasQuest(self.quest_on_die)
    if q and not q:isEnded() then q:check('on_quest_mon_die', self, src) end
  end
  if self.achievement_on_die then
    world:gainAchievement(self.achievement_on_die, game.player, self)
  end

  if self.on_die_post then self:on_die_post(src) end
  return true
end

function _M:levelupMsg(down)
  -- Subclasses handle the actual mechanics of leveling up; here we just
  -- print the message and add the flyer.
  local stale = false
  if self.level_hiwater then
    stale = self.level_hiwater >= self.level
    self.level_hiwater = math.max(self.level_hiwater, self.level)
  end

  -- If actor requested silent levelup, punt.
  if self.silent_levelup then return not down and not stale end

  local color = down and '#FF6666#' or '#00FFFF#'
  local gains = down and 'loses' or stale and 'regains' or 'gains'
  game.logSeen(self, "%s%s %s level %d.#LAST#", color, self.name, gains, self.level)
  if self.x and self.y and game.level.map.seens(self.x, self.y) then
    self:addFlyer(80, 0.5, -2, down and 'LEVEL DOWN!' or 'LEVEL UP!', down and {255, 96, 96} or stale and {255, 0, 255} or {0, 255, 255})
  end

  -- Return true if this is the first time we've hit this level.
  return not down and not stale
end

function _M:attack(target)
  self:bumpInto(target)
end

function _M:adjGold(delta)
  self.gold = math.max(0, self.gold + delta)
  self.changed = true
  if self.player then
    world:gainAchievement('GOLD_100K', self)
    world:gainAchievement('GOLD_500K', self)
    world:gainAchievement('GOLD_1M', self)
  end
end

function _M:getArmor()
  return self.ac + (self.to_a or 0)
end

function _M:getTalentResourceCost(t, rsrc)
  if t[rsrc] then return util.getval(t[rsrc], self, t) end
  if self.player and t.player and t.player[rsrc] then
    return util.getval(t.player[rsrc], self, t)
  end
  return nil
end

--- Called before a talent is used
-- Check the actor can cast it
-- @param ab the talent (not the id, the table)
-- @return true to continue, false to stop
function _M:preUseTalent(ab, silent, dry_run)
  if not self:enoughEnergy() then print("fail energy") return false end
  local no_fail = self:attr('force_talent_no_fail')
  local no_rsrc = self:attr('force_talent_ignore_ressources')

  if ab.god and (not self.isFollowing or not self:isFollowing(ab.god)) then
    if not silent then
      game.logPlayer(self, 'You are not a follower of %s.', ab.god)
    end
    return false
  end

  if ab.is_spell and ab.skill and not spell.spell_skill_level_ok(self, ab) and not no_fail then
    local Skills = require 'mod.class.interface.PlayerSkills'
    local SK = Skills.skills_def
    local sk = {}
    for _, tag in ipairs(type(ab.skill) == 'table' and ab.skill or { ab.skill }) do
      sk[#sk+1] = SK[tag] and SK[tag].name or '???('..tag..')'
    end
    sk = table.concat(sk, '/')
    if not silent then
      game.logPlayer(self, 'You do not have enough %s skill to cast %s.', sk, ab.name)
    end
    return false
  end

  if ab.base_fail and not no_fail then
    -- By convenient coincidence, all the talents that use T2's equivalent
    -- of spell.fail_prob() also have these three checks.
    if self:has('ANTIMAGIC_FIELD') then
      if not silent then
	game.logPlayer(self, 'Your anti-magic field disrupts any magic attempts.')
      end
      return false
    end
    if self:has('NO_MAGIC') then
      if not silent then
	game.logPlayer(self, 'Your anti-magic shell disrupts any magic attempts.')
      end
      return false
    end
    if self:has('CONFUSED') then
      if not silent then
	game.logPlayer(self, 'You are too confused!')
      end
      return false
    end
  end

  if ab.power then
    if self.level < ab.power.level then
      if not silent then
	game.logPlayer(self, 'You need to attain level %d to use this power.', ab.power.level)
      end
      return false
    end
    if self:has('CONFUSED') then
      if not silent then
	game.logPlayer(self, 'You are too confused to use this power.')
      end
      return false
    end
  end

  if ab.mode == "sustained" then
    -- TODO We don't yet have any sustained talents that cost anything.
    --[[
    if ab.sustain_power and self.max_power < ab.sustain_power and not self:isTalentActive(ab.id) then
      game.logPlayer(self, "You do not have enough power to activate %s.", ab.name)
      return false
    end
    ]]
    if not no_rsrc and not self:isTalentActive(ab.id) then
      local p = { no_use=false }
      self:fireCallbacks('on_use_talent_pre_rsrc', ab, silent, dry_run, p)
      if p.no_use then return false end
    end
  else
    if not no_rsrc then
      -- Slight Hack(TM):  NPCs don't have mana, and don't use it to cast
      -- their spells.
      -- Slight Hack(TM):  Talents with a power{} field can be cast from
      -- hit points.
      local mana_cost = self:getTalentResourceCost(ab, 'mana')
      if mana_cost and mana_cost > (self.mana or 0) then
	if not ab.power or mana_cost >= self.life then
	  -- TODO Allow the player to try anyway, with bad effects.
	  if self.player and not silent then
	    game.logPlayer(self, "You do not have enough mana%s to cast %s.", (ab.power and ' or life' or ''), ab.name)
	  end
	  return false
	elseif not dry_run and not silent and self.player then
	  -- Prompt for this.
	  local function cb(ok) self:talentDialogReturn(ok) end
	  local msg = [[You do not have enough mana to use this power, but you can use it anyway at the cost of %d life.  Proceed?]]
	  msg = msg:format(mana_cost)
	  local ret = self:talentDialog(ui_util.yesnoLongPopup(ab.name, msg, 600, cb))
	  if not ret then return false end
	end
      end
      local piety_cost = self:getTalentResourceCost(ab, 'piety')
      if piety_cost and piety_cost > (self.piety or 0) and
	-- Slight Hack(TM):  NPCs don't have piety.
	 (self.player or not ab.type:match('^npc/') or not ab.no_npc_piety)
      then
	if self.player and not silent then
	  game.logPlayer(self, "You do not have enough piety to cast %s.", ab.name)
	end
	return false
      end
      local p = { no_use=false }
      self:fireCallbacks('on_use_talent_pre_rsrc', ab, silent, dry_run, p)
      if p.no_use then return false end
    end
  end

  -- Talent-specific checks
  if ab.on_pre_use and
     not (ab.mode == 'sustained' and self:isTalentActive(ab.id)) and
     not ab.on_pre_use(self, ab, silent, dry_run) and not no_fail
  then
    return false
  end
  -- Temporary Hack(TM) to prevent anyone from trying to use one of the
  -- unimplemented spells.
  if ab.info_aux then
    if not silent then
      game.logSeen(self, 'Sorry, not implemented (%s)', ab.name)
    end
    return false
  end

  -- Random failure chance.  Don't check this in dry-run mode, since that's
  -- used to determine how to highlight talents in the HotkeysDisplay.
  if not no_fail and ab.base_fail and not dry_run then
    local fail_prob = spell.fail_prob(self, ab)
    if rng.percent(fail_prob) then
      if not silent then
	if not self.player then
	  local name = self:getName{def_art=true, check_seen=true}
	  local use = (ab.is_spell or ab.god) and 'cast' or 'use'
	  game.logSeen(self, '%s tries to %s %s, but fails.', name, use, ab.name)
	else
	  game.logPlayer(self, 'You failed to concentrate hard enough!')
	end
      end
      if ab.on_fail then ab.on_fail(self, ab, fail_prob) end
      -- This talent failure mode uses energy and mana/piety.
      self:useEnergy()
      if not no_rsrc then
	if self.player and ab.mana then
	  self:incMana(-util.getval(ab.mana, self, ab))
	end
	if ab.piety then
	  self:incPiety(-util.getval(ab.piety, self, ab))
	end
      end
      return false
    end
  end

  -- NPC spells can be blocked by player antimagic.
  if not no_fail and not dry_run and not self.player and ab.is_spell and game.player:has('ANTIMAGIC_FIELD') then
    local dist = core.fov.distance(self.x, self.y, game.player.x, game.player.y)
    print(('[ANTIMAGIC] dist %d vs. rad %d, power %d%%'):format(dist, game.player.antimagic_rad, game.player.antimagic))
    if dist <= (game.player.antimagic_rad or 0) and rng.percent(game.player.antimagic) then
      if not silent then
	game.log([[Your anti-magic field disrupts %s's spell.]], self:getName{def_art=true, check_seen=true})
      end
      return false
    end
  end

  -- Failure chance for race/class/object powers.  As above, we skip this
  -- in dry-run mode.
  if not no_fail and ab.power and not dry_run then
    local diff = ab.power.diff
    local eff = self:hasEffect(self.EFF_STUN)
    if eff then
      diff = diff + eff.dur
    elseif self.level > ab.power.level then
      diff = diff - math.min(math.floor((self.level - ab.power.level)/3), 10)
    end
    diff = math.max(diff, 5)
    -- Annoying Hack(TM):  The failure chance in this mode apparently uses
    -- the "raw" stat, with no adjusters, encode a la T2 (with ie. 18/100
    -- corresponding to a value of 118).
    -- [cf. src/powers.c function power_chance()]
    local stat = self:getStat(ab.power.stat, {raw='cur', no_birth=true, no_equip=true, no_temp=true})
    stat = stat <= 150 and math.floor(stat/10) + 3 or stat - 132
    local d2 = math.floor(diff/2)
    if rng.range(1, stat) < d2 + rng.range(1, d2) then
      if not silent then
	-- Only players should be getting to this point.
	game.logPlayer(self, 'You failed to concentrate hard enough!')
      end
      if ab.on_fail then ab.on_fail(self, ab, fail_prob) end
      -- This talent failure mode uses energy and mana/piety.
      self:useEnergy()
      if not no_rsrc then
	if self.player and ab.mana then
	  local mana_cost = util.getval(ab.mana, self, ab)
	  if self.mana >= mana_cost then
	    self:incMana(-mana_cost)
	  else
	    self:takeHit(mana_cost, self, 'concentrating too hard')
	  end
	end
      end
      return false
    end
  end

  local p = { no_use=false }
  self:fireCallbacks('on_use_talent_pre', ab, silent, dry_run, p)
  if p.no_use then return false end

  -- Our work here... is done.
  return true
end

function _M:logTalentMessage(ab)
  local name = self:getName{def_art=true, check_seen=true}
  -- Allow for silent talents
  if ab.message == false then return end
  if ab.message then
    game.logSeen(self, "%s", self:useTalentMessage(ab))
  elseif ab.mode == "sustained" and not self:isTalentActive(ab.id) then
    game.logSeen(self, "%s activates %s.", name, ab.name)
  elseif ab.mode == "sustained" and self:isTalentActive(ab.id) then
    game.logSeen(self, "%s deactivates %s.", name, ab.name)
  elseif ab.is_spell or ab.god then
    game.logSeen(self, "%s casts %s.", name, ab.name)
  else
    game.logSeen(self, "%s uses %s.", name, ab.name)
  end
end

-- Modified from the parent to split out substitutions into a separately
-- usable method.
function _M:useTalentMessage(ab)
  if not ab.message then return nil end
  local str = util.getval(ab.message, self, ab)
  local _, _, target = self:getTarget()
  return self:convertTalentMessage(target, str)
end

function _M:convertTalentMessage(target, msg)
  local function actor_name(a)
    return (a and a.getName and a:getName{def_art=true, check_seen=not a.player}) or (a and a.name) or 'something'
  end

  local pronouns = {
    subject='it', object='it', possessive='its', reflexive='itself',
  }

  if not target then _, _, target = self:getTarget() end
  local subs = { source=actor_name(self), target=actor_name(target) }
  for prn, dflt in pairs(pronouns) do
    subs['source.pronoun.'..prn] = self:pronoun(prn)
    subs['target.pronoun.'..prn] = target and target.pronoun and target:pronoun(prn) or dflt
  end
  for key, repl in pairs(subs) do
    msg = msg:gsub('@'..key..'@', repl):gsub('@'..key:capitalize()..'@', repl:capitalize())
  end
  return msg
end

--- Called after a talent is used
-- Consume a turn, mana, stamina, etc. as appropriate
-- @param ab the talent (not the id, the table)
-- @param ret the return of the talent action
-- @return true to continue, false to stop, table to continue with special info
function _M:postUseTalent(ab, ret)
  if not ret then return end

  if not util.getval(ab.no_energy, self, ab) then
    local energy = self:energyForTalent(ab)
    if type(ret) == 'table' and ret.use_speed then
      energy = game.energy_to_act * ret.use_speed
    end
    self:useEnergy(energy)
  end

  if not self:attr("force_talent_ignore_ressources") then
    if ab.mode == "sustained" then
      if not self:isTalentActive(ab.id) then
	-- TODO We don't yet have any sustained talents that cost anything.
	--[[
	if ab.sustain_power then
	self.max_power = self.max_power - ab.sustain_power
	end
	]]
	self:fireCallbacks('on_use_talent_post_rsrc', ab, ret)
	self:installCallbacks(ab, 'active_callbacks', ab.id, 'talent')
      else
	-- TODO As above.
	--[[
	if ab.sustain_power then
	self.max_power = self.max_power + ab.sustain_power
	end
	]]
	self:fireCallbacks('on_use_talent_post_rsrc', ab, ret)
	self:uninstallCallbacks(ab, 'active_callbacks', ab.id, 'talent')
      end
    else
      -- Slight Hack(TM):  NPCs don't have mana, and don't use it to cast
      -- their spells.
      local mana_cost = util.getval(ab.mana, self, ab)
      if mana_cost and self.mana then
	if ab.power and self.mana < mana_cost then
	  self:takeHit(mana_cost, self, 'concentrating too hard')
	else
	  self:incMana(-mana_cost)
	end
      end
      local piety_cost = util.getval(ab.piety, self, ab)
      if piety_cost and ab.piety then
	self:incPiety(-piety_cost)
      end
      self:fireCallbacks('on_use_talent_post_rsrc', ab, ret)
    end
  end

  return true
end

function _M:energyForTalent(t)
  local use_speed = 1
  if t.use_speed then
    use_speed = util.getval(t.use_speed, self, t)
  end
  if t.is_spell and self:has('fast_spellcast') then
    -- Mage staffs reduce spellcasting time to 80% normal.
    use_speed = use_speed * 0.8
  end
  return game.energy_to_act * use_speed
end

function _M:forceUseTalent(t, def)
  if def.force_no_fail then self:attr('force_talent_no_fail', 1) end
  local ret = { engine.interface.ActorTalents.forceUseTalent(self, t, def) }
  if def.force_no_fail then self:attr('force_talent_no_fail', -1) end
  return unpack(ret)
end

--- Return the full description of a talent
-- You may overload it to add more data (like power usage, ...)
function _M:getTalentFullDescription(t)
  local d = {}

  if t.mode == "passive"
    then d[#d+1] = "#6fff83#Use mode: #00FF00#Passive"
  elseif t.mode == "sustained" then
    d[#d+1] = "#6fff83#Use mode: #00FF00#Sustained"
  else
    d[#d+1] = "#6fff83#Use mode: #00FF00#Activated"
  end

  if spell.stick.source then
    -- Slight Hack(TM):  When describing a wand/staff, use Magic-Device.
    d[#d+1] = '#6fff83#Skill: #GOLD#Magic-Device'
  elseif t.skill then
    local Skills = require 'mod.class.interface.PlayerSkills'
    local SK = Skills.skills_def
    local s = type(t.skill) == 'table' and 's' or ''
    local sk = {}
    for i, id in ipairs(type(t.skill) == 'table' and t.skill or { t.skill }) do
      sk[i] = SK[id] and SK[id].name or '???('..id..')'
    end
    d[#d+1] = '#6fff83#Skill'..s..': #GOLD#'..table.concat(sk, ', ')
    -- TODO Maybe describe Sorcery/Prayer substitutions.
  end

  if t.is_spell and t.skill_level then
    -- This will transparently use wand/staff data if present.
    local eff = spell.get_spell_level(self, t, 50, 0)
    if not eff then
      d[#d+1] = '#FF0000#??? Spell level info missing#FFFFFF'
    elseif eff < 1 then
      d[#d+1] = ('#6fbf73#Req. skill level: #FFFFFF#%d'):format(t.skill_level)
    else
      d[#d+1] = ('#6fff83#Spell level: #FFFFFF#%d'):format(eff)
    end
  end

  if t.power then
    local diff = t.power.diff
    local eff = self:hasEffect(self.EFF_STUN)
    if eff then
      diff = diff + eff.dur
    elseif self.level > t.power.level then
      diff = diff - math.min(math.floor((self.level - t.power.level)/3), 10)
    end
    local d2 = math.floor(diff/2) + math.floor(diff/2)/2
    -- Annoying Hack(TM), redux:  cf. preUseTalent() above.
    local stat = self:getStat(t.power.stat, {raw='cur', no_birth=true, no_equip=true, no_temp=true})
    stat = stat <= 150 and math.floor(stat/10) + 3 or stat - 132
    local fail = math.floor(d2/stat * 100)

    d[#d+1] = ('#6fbf73#Req. level: #FFFFFF#%d'):format(t.power.level)
    d[#d+1] = ('#6fbf73#Failure chance: #FFFFFF#%d%% #6fbf73#(decreases with level and %s)#FFFFFF#'):format(fail, t.power.stat)
  end

  if t.god then
    d[#d+1] = ('#6fff83#God: #FFFFFF#%s'):format(t.god)
  end

  if t.base_fail then
    -- This will transparently use wand/staff data if present.
    local l = ('#6fbf73#Failure chance: #FFFFFF#%d%%'):format(spell.fail_prob(self, t))
    if t.skill_level or t.stat then
      local dd = {}
      if t.stat then dd[#dd+1] = t.stat end
      if t.skill_level then dd[#dd+1] = 'skill level' end
      l = l .. ' #6fbf73#(decreases with ' .. table.concat(dd, ' and ') .. ')#FFFFFF#'
    end
    d[#d+1] = l
  end

  if not spell.stick.source then
    if t.mana then
      d[#d+1] = ('#6fff83#Mana cost: #FFFFFF#%d'):format(util.getval(t.mana, self, t))
      if t.power then
	d[#d+1] = '#{italic}##6fbf73#[Can be cast from life if mana is too low]#LAST##{normal}#'
      end
    end
    if t.piety then
      d[#d+1] = ('#6fff83#Piety cost: #FFFFFF#%d'):format(util.getval(t.piety, self, t))
    end
    self:fireCallbacks('on_talent_desc_rsrc', t, d)
  end

  if self:getTalentRange(t) > 1 then
    d[#d+1] = "#6fff83#Range: #FFFFFF#"..self:getTalentRange(t)
  else
    d[#d+1] = "#6fff83#Range: #FFFFFF#melee/personal"
  end
  if t.cooldown and not spell.stick.source then
    d[#d+1] = "#6fff83#Cooldown: #FFFFFF#"..t.cooldown
  end

  local uspeed = 'Full Turn'
  if util.getval(t.no_energy, self, t) then
    uspeed = "Instant (#LIGHT_GREEN#0%#LAST# of a turn"
  else
    local spd = (self:energyForTalent(t) * 100)/game.energy_to_act
    local typ = t.is_spell and 'Spell' or 'Normal'
    uspeed = ("%s (#LIGHT_GREEN#%d%%#LAST# of a turn)"):format(typ, spd)
  end
  d[#d+1] = '#6fff83#Usage Speed:#FFFFFF# '..uspeed

  if t.inertia and not spell.stick.source then
    d[#d+1] = ('#6fbf73#If used with Inertia Control: #FFFFFF#Flow power %d, recasts every %d turns'):format(t.inertia[1], t.inertia[2])
  end

  local info = util.getval(t.info, self, t)
  d = table.concat(d, "\n").."\n#6fff83#Description: #FFFFFF#"..info
  if t.is_spell and t.info_aux then
    d = d .. '\n\n#YELLOW#This spell is not yet implemented.  We Apologize for the Inconvenience.™#LAST#'
  end
  return d:toTString()
end

-- Apply effects when hit in melee.
function _M:onHitInMelee(attacker, wpn)
  -- Retaliatory damage auras.
  for rtype, method in pairs(self.melee_retal or {}) do
    local rdt = DamageType:get(rtype)
    local radj = rdt.aura_adj or rdt.name
    local dam = { atk_msg = "'s " .. radj .. ' aura hits %s' }
    if method == 'mondam' then
      local dd, ds = 1 + math.floor(self.level/26), 1 + math.floor(self.level/17)
      dam[1] = rng.dice(dd, ds)
    elseif method == 'plyrdam' then
      dam[1] = rng.dice(2, 6)
    end
    rdt.projector(self, attacker.x, attacker.y, rtype, dam)
    if game.player:hasLOS(self.x, self.y) and game.player:canSee(self) then
      self:learnFlags('melee_retal', rtype)
    end
    if attacker.dead then return end
  end

  if not self.dead then
    self:fireCallbacks('on_hit_in_melee', attacker, wpn)
  end
end

--- How much experience is this actor worth
-- @param target to whom is the exp rewarded
-- @return the experience rewarded
function _M:worthExp(target)
  -- TODO Don't get experience from killing friendlies.
  return (self.exp_worth * self.level)/math.max(target.level, 1)
end

--- Can the actor see the target actor
-- This does not check LOS or such, only the actual ability to see it.<br/>
-- Check for telepathy, invisibility, stealth, ...
function _M:canSeeNoCache(actor, def, def_pct, t)
  t = t or {}
  if not actor then return false, 0 end
  -- We can always see ourself.  Downstream code will take care of making
  -- us "look" invisible as needed.
  if not t.unsee_hack and actor == self then return true, 100 end

  if self.esp then
    local esp = false
    for f, _ in pairs(self.esp) do
      esp = esp or f == 'ALL' or actor:has(f)
    end
    if esp then
      -- Careful; without the force flag here, ActorLearn:learnFlags() will
      -- call Actor:canSee(), which could get us into an infinite loop.
      if self.player and actor.learnFlags then
	actor:learnFlags(true, 'flags', 'EMPTY_MIND')
	actor:learnFlags(true, 'flags', 'WEIRD_MIND')
      end
      if not actor:has('EMPTY_MIND') and (not actor:has('WEIRD_MIND') or rng.percent(10)) then
	return true, 100
      end
    end
  end

  -- Slight Hack(TM):  We can sense anything that's aggravating us.
  if actor:has('AGGRAVATE') then return true, 100 end

  -- If we're blind, we can't see anything.
  if self:has('BLIND') then return false, 0 end

  -- If the target is invisible, we can only see them if we can see through
  -- their invisibility.
  if actor:has('INVISIBLE') and not self:canSeeInvis(actor) then
    return false, 0
  end

  if def ~= nil then
    return def, def_pct
  else
    return true, 100
  end
end

--- Can the actor see the target actor
-- Work from cache or falls through to :canSeeNoCache() above.
function _M:canSee(actor, def, def_pct)
  if not actor then return false, 0 end

  self.can_see_cache = self.can_see_cache or {}
  local key = tostring(def) .. '/' .. tostring(def_pct)
  local scsc = self.can_see_cache

  if scsc[actor] and scsc[actor][key] then
    -- Cache hit.
    local hit = scsc[actor][key]
    return hit[1], hit[2]
  end

  -- Cache miss, populate cache.
  local ret, pct = self:canSeeNoCache(actor, def, def_pct)
  scsc[actor] = scsc[actor] or {}
  scsc[actor][key] = { ret, pct }

  -- Not sure what this does; the T4 code describes at as 'Make sure the
  -- display updates'.
  if self.player and type(def) == 'nil' and actor._mo then
    actor._mo:onSeen(ret)
  end

  return ret, pct
end

-- Reset our :canSee() cache.
function _M:resetCanSeeCache()
  self.can_see_cache = {}
  setmetatable(self.can_see_cache, {__mode="k"})
end

-- Reset the :canSee() cache of anyone who can see us.
function _M:resetViewersCanSeeCache()
  if game.level then
    for _, e in pairs(game.level.entities) do
      if e.can_see_cache and e.can_see_cache[self] then
	e.can_see_cache[self] = nil
      end
    end
    game.level.map:updateMap(self.x, self.y)
  end
end

-- Overridden by Player and NPC.
function _M:canSeeInvis(target)
  return false
end

-- Attempt to open the provided door.  Returns true if successful.  If
-- 'dry_run' is set, just returns true if we could possibly open the door.
-- Overridden in NPC and Player
-- cf. Grid:block_move()
function _M:tryOpenDoor(g, x, y, dry_run)
  return false
end

function _M:computeGlobalSpeed()
  if self.speed < 0 then
    self.global_speed = 1/(1 - self.speed/10)
  elseif self.speed <= 26 then
    self.global_speed = 1 + self.speed/10
  elseif self.speed <= 34 then
    -- Everything from here down is an asymptotic approach to 500% speed,
    -- roughly mimicking ToME2's tables.c:extract_energy[] table.  Split
    -- out into cases, as there was no obvious function that suitably
    -- reproduced the table; this only gets called by the resolver, though
    -- (and when actor speed actually changes), so this shouldn't be too
    -- expensive.
    self.global_speed = 3.6 + (self.speed - 26)/20
  elseif self.speed <= 46 then
    self.global_speed = 4 + (self.speed - 34)/30
  elseif self.speed <= 50 then
    self.global_speed = 4.4 + (self.speed - 46)/40
  elseif self.speed <= 70 then
    self.global_speed = 4.5 + (self.speed - 50)/50
  else
    self.global_speed = 5 - 1/(self.speed - 60)
  end
end

function _M:sortInven(inven)
  if not inven then inven = self.inven[self.INVEN_INVEN] end
  inven = self:getInven(inven)
  if not inven then return end
  obj_util.sort_inven(inven)
  self.changed = true
end

function _M:added()
  engine.Entity.added(self)
  if self.on_add then self:on_add() end
end

-- Pval-based fields that don't always have a conveyed flag associated with
-- them.
-- Don't know if I'll need any more of these, but just in case...
local pval_to_has = {
  INVISIBLE = 'invisibility',
}

function _M:has(flag)
  local pv = pval_to_has[flag]
  if pv and type(self[pv]) == 'number' and self[pv] > 0 then
    return true
  end
  return (type(self.flags) == 'table' and self.flags[flag]) or
	 self:hasEffect(self['EFF_'..flag]) or
	 self:isTalentActive(self['T_'..flag]) or
	 (type(self.abilities) == 'table' and self.abilities[flag]) or
	 (type(self[flag]) == 'number' and self[flag] ~= 0)
end

function _M:hasTweakableVisuals()
  return self:has('ATTR_MULTI') or self:has('ATTR_CLEAR') or self:has('CHAR_CLEAR')
end

-- Overridden in NPC and Player.
function _M:resistDam(damtype)
  return false
end

-- Overridden in NPC and Player.
function _M:immuneDam(damtype)
  return false
end

-- Overridden in NPC and Player.
function _M:vulnDam(damtype)
  return false
end

-- Copied from parent method (q.v.) to avoid error in case 'ab' is null.
-- Which is done by :talentDialog()'s .unload() method in the parent, so it
-- should really be expecting that...
function _M:onTalentLuaError(ab, err)
  if self.talent_error then return end -- handle only the first error
  self.talent_error = {
    Actor = self,
    uid = self.uid,
    name = self.name,
    err = err,
    x = self.x,
    y = self.y,
    turn = game.turn,
  }
  if ab then self.talent_error[ab.id] = ab end
  print("##Use Talent Lua Error##", ab and ab.id, "Actor:", self.uid, self.name)
  _M._talent_errors = _M._talent_errors or {} -- log the error globally
  table.insert(_M._talent_errors, self.talent_error)
  return
end

-- Copied from parent method (q.v.) to add support for an :on_error()
-- method that will be called on a dialog if an error occurs.  We still
-- call :onTalentLuaError(nil, err), on the theory that it must be there
-- for a reason...
function _M:talentDialog(d)
  if not game:hasDialog(d) then game:registerDialog(d) end

  local dialog_returns_list, dialog_returns = self:talentDialogData()
  dialog_returns_list[#dialog_returns_list+1] = d

  local co = coroutine.running()
  d.unload = function(dialog)
    local ok, err = coroutine.resume(co, dialog_returns[d])
    if not ok and err then
      print(debug.traceback(co))
      if d.on_error then d:on_error() end
      self:onTalentLuaError(nil, err)
      error(err)
    end
  end
  local ret = coroutine.yield()

  dialog_returns[d] = nil
  table.removeFromList(dialog_returns_list, d)

  return unpack(ret or {})
end
