-- $Id: Combat.lua 4256 2021-08-02 15:02:20Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require "engine.class"
local DamageType = require "engine.DamageType"
local Target = require "engine.Target"

--- Interface to add ToME combat system
module(..., package.seeall, class.make)

--- Checks what to do with the target
-- Talk ? attack ? displace ?
function _M:bumpInto(target)
  local reaction = self:reactionToward(target)
  if reaction < 0 then
    return self:attackTarget(target)
  elseif reaction >= 0 then
    if self.player and target.on_bump then
      target:on_bump(self)
    elseif self.player and target.can_talk then
      local Chat = require 'engine.Chat'
      local chat = Chat.new(target.can_talk, target, self, {npc=target, player=self})
      chat:invoke()
      if target.can_talk_only_once then target.can_talk = nil end
    elseif target.player and self.can_talk then
      local Chat = require 'engine.Chat'
      local chat = Chat.new(self.can_talk, self, target, {npc=self, player=target})
      chat:invoke()
      if self.can_talk_only_once then self.can_talk = nil end
    elseif self.player then
      game.logPlayer(self, '%s is in your way.', target:getName{def_art=true, check_seen=true}:capitalize())
    end
  end
end

function _M:checkHit(target, power)
  -- Hit or miss, attacking an NPC wakes it up.
  if target.ai_state and target.ai_state.sleep then
    target.ai_state.sleep = 0
  end

  print('[COMBAT] checkHit('..power..')')
  -- Fixed 5% miss chance and 5% hit chance.
  local edge = rng.range(1, 20)
  if edge <= 2 then
    print('[COMBAT] checkHit:  edge case 5% '..(edge == 1 and 'hit' or 'miss'))
    return edge == 1
  end
  if power <= 0 then
    print(('[COMBAT] checkHit:  auto-miss on power=%d'):format(power))
    return false
  end

  if not self:canSee(target) then
    power = math.floor((power+1)/2)
    print(('[COMBAT] checkHit:  power reduced to %d for invisible target'):format(power))
  end

  if self.player then
    print('[COMBAT] checkHit:  attacker is player')
  elseif target.player then
    print('[COMBAT] checkHit:  target is player')
  end
  local tgt_ac = target:getArmor()
  local atk_pow = power
  -- Slight Hack(TM):  We also need to use use player-based calculations
  -- for player-set traps [cf. trap_util.set_trap()]
  if self.player or self.setter then
    print(('[COMBAT] check_hit:  target AC %d, luck adj %d'):format(tgt_ac, game.player:getLuckScale(-10, 10)))
    atk_pow = atk_pow + game.player:getLuckScale(-10, 10)
  else
    atk_pow = atk_pow + 3*math.min(self.level, 1)
    if target.player then
      atk_pow = atk_pow - target:getLuckScale(-10, 10)
    end
  end
  local tgt_ac_adj = math.floor((tgt_ac*3)/4)
  local k = rng.range(0, atk_pow-1)
  if k < tgt_ac_adj then
    print(('[COMBAT] checkHit:  0-%d -> %d < %d -> miss'):format(atk_pow-1, k, tgt_ac_adj))
    return false
  else
    print(('[COMBAT] checkHit:  0-%d -> %d >= %d -> hit'):format(atk_pow-1, k, tgt_ac_adj))
    return true
  end
end

-- cf. tot_dam_aux() in src/cmd1.c
function _M:compute_brand_slay_mult(target, wpn)
  local mults = {}
  for f, _ in pairs(wpn.slay_flags or {}) do
    if f:find('^X_') then
      mults[#mults+1] = { flag=f:gsub('^X_', ''), 5 }
    else
      local m = (f == 'ANIMAL' or f == 'EVIL') and 2 or 3
      mults[#mults+1] = { flag=f, m }
    end
  end
  for dt, _ in pairs(wpn.brand_dam or {}) do
    mults[#mults+1] = { damtype=dt, 3, 6 }
  end
  self:fireCallbacks('on_wpn_slay_brand', target, wpn, mults)

  local mult, do_pois = 1, false
  for _, def in ipairs(mults) do
    if type(def) == 'function' then
      mult = math.max(mult, def(self, target, wpn))
    elseif type(def) == 'table' and def.flag and target:has(def.flag) then
      if target.learnFlags then target:learnFlags('flags', ff) end
      mult = math.max(mult, def[1])
    elseif type(def) == 'table' and def.damtype then
      local vuln = target:vulnDam(def.damtype, {learn=true})
      local res = target:resistDam(def.damtype, {learn=true, or_immune=true})
      mult = math.max(mult, vuln and def[2] or not res and def[1] or 1)
      do_pois = do_pois or def.damtype == DamageType.POISON
    end
  end

  if do_pois then
    -- Special poison effects (cf. src/cmd1.c lines 476-491)
    local vuln = target:vulnDam(DamageType.POISON, {learn=true})
    local res = target:resistDam(DamageType.POISON, {learn=true, or_immune=true})
    local doit = vuln and rng.percent(95) or (rng.percent(50) and not rng.percent(target.level))
    if doit then
      target:setEffect(target.EFF_POISON, vuln and base*2 or base,
		       {merge='combine', src=self, combine_msg='@Target@ is more poisoned.'})
    end
  end

  return mult
end


-- cf. critical_shot() in src/cmd1.c
function _M:criticalShot(ammo, dam, skill)
  -- I thinks player traps are supposed to use player skill/luck.
  local src = self.resolveSource and self:resolveSource() or self
  local i = ammo.weight*10 + ((self.to_h or 0) + ammo.to_h)*4 + 50*(self.extra_crit or 0)
  if src.getSkillLevel then i = i + src:getSkillLevel(skill, 100) end
  if src.getLuckScale then i = i + src:getLuckScale(-100, 100) end

  if rng.range(1, 5000) <= i then
    local k = ammo.weight*10 + rng.range(1, 500)
    -- TODO should probably only log if we can see the target.
    if k < 500 then
      game.log('It was a good hit!')
      dam = 2*dam + 5
    elseif k < 1000 then
      game.log('It was a great hit!')
      dam = 2*dam + 10
    else
      game.log('It was a superb hit!')
      dam = 3*dam + 15
    end
  end
  return dam
end

-- cf. do_cmd_fire() in src/cmd2.c
function _M:ammoDam(ammo, target, launcher)
  if not launcher then
    local l_inven = self:getInven(self.INVEN_LAUNCHER)
    launcher = l_inven[1]
  end

  local dam = rng.dice(ammo.combat[1], ammo.combat[2]) + ammo.to_d + launcher.to_d
  local mul = launcher:ammoMult() + self.extra_might
  dam = dam*mul + self.to_d_ranged
  dam = dam * self:compute_brand_slay_mult(target, ammo)
  dam = self:criticalShot(ammo, dam, 'ARCHERY')
  return math.max(dam, 0)
end

-- Handle ammo explosion.  [cf. do_cmd_fire() in src/cmd2.c]
function _M:ammoExplode(ammo, x, y)
  -- Quick sanity check.
  if not ammo.explode_dam_type then return end
  local ammo_name = ammo:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
  local src_name = self:getName{def_art=true, check_seen=true, show_setter=true}
  local dam = {
    math.floor(2*ammo.explode_mult*(rng.dice(ammo.combat[1], ammo.combat[2]) + (ammo.to_d or 0))),
    display_src = { name=([[%s's exploding %s]]):format(src_name, ammo_name) },
  }
  spell.ball_spell(self, nil, ammo.explode_dam_type, dam, ammo.explode_r, x, y, 0, spell.simple_blast_particle)
end

-- Drop fired ammo.  We mostly reproduce the tail end of Actor:dropFloor()
-- [q.v.]
function _M:ammoDropFloor(ammo, x, y)
  local ammo_name = ammo:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
  local dx, dy, long = util.drop_loc_near(x, y, 10, ammo.unique)
  local src_name = self:getName{def_art=true, check_seen=true, show_setter=true}
  if not dx or not dy then
    -- Very bad.
    game.logSeenXY(x, y, [[%s's %s spontaneously disintegrates!]], src_name:capitalize(), ammo_name)
    -- If this was an artifact, be nice and put it back into circulation.
    ammo:removed()
  else
    if long then
      game.logSeenXY(x, y, [[%s's %s vanishes!]], src_name:capitalize(), ammo_name)
    end
    -- Slight Hack(TM):  Set a temporary auto-pickup flag on this ammo.
    -- [cf. Player:on_pickup_object]
    ammo.auto_pickup = 'temp'
    local ok, idx = game.level.map:addObject(dx, dy, ammo)
    if ok and game.level.map.attrs(dx, dy, "on_drop") then
      game.level.map.attrs(dx, dy, "on_drop")(ammo, dx, dy, idx, ammo)
    end
  end
end
