-- $Id: AlchemyIngredientsDialog.lua 3041 2017-08-30 02:40:05Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local List = require 'engine.ui.List'
local Button = require 'engine.ui.Button'
local Textzone = require 'engine.ui.Textzone'
local TextzoneList = require 'engine.ui.TextzoneList'
local Separator = require 'engine.ui.Separator'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(actor, recipe, n, callback)
  self.actor = actor
  self.recipe = recipe
  self.needed = n
  self.callback = callback
  self.choices = { n=0 }

  local s = n == 1 and '' or 's'
  local title = 'Ingredient%s for:  %s'
  Dialog.init(self, title:format(s, recipe.name), 1, 1)

  local ingr_desc = util.alchemy.ingredient_simple_desc(recipe, n)
  local instructions = [[#SLATE#Choose the %s to use as the ingredient%s for the artifact power'%s'.  Click on items in the top list to move them one at a time into your list of choices.  Click on items in the bottom list to move them back into the candidates list.#LAST#]]
  self.c_desc = Textzone.new {
    width = 600,
    auto_height = true,
    no_color_bleed = true,
    text = instructions:format(ingr_desc, s, recipe.name),
  }

  self.c_count = TextzoneList.new {
    width = 500,
    height = self.font_h,
  }
  self.c_count:generate()
  self:updateCount()

  self.c_ok = Button.new {
    text = 'Accept',
    fct = function() self:finish() end,
  }
  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function() self.key:triggerVirtual('EXIT') end,
  }

  self:generateLists()
  self.c_above = List.new {
    width = 600,
    nb_items = util.bound(#self.cands_list, 3, 8),
    list = self.cands_list,
    scrollbar = true,
    fct = function(item, sel) self:selectItem(item, true) end,
    select = function(item, sel) end,
  }
  self.c_below = List.new {
    width = 600,
    nb_items = util.bound(#self.cands_list, 3, 8),
    list = self.choices_list,
    scrollbar = true,
    fct = function(item, sel) self:selectItem(item, false) end,
    select = function(item, sel) end,
  }

  local c_avail = Textzone.new {
    auto_width = true,
    auto_height = true,
    no_color_bleed = true,
    text = '#{bold}#Available#{normal}#',
  }
  local c_chosen = Textzone.new {
    auto_width = true,
    auto_height = true,
    no_color_bleed = true,
    text = '#{bold}#Chosen#{normal}#',
  }
  c_avail:generate()
  c_chosen:generate()
  local vs1 = Separator.new { dir='vertical', size=590 }
  local vs2 = Separator.new { dir='vertical', size=590 }
  self:loadUI {
    { left=0, top=0, ui=self.c_desc },
    { left=20, top=self.c_desc, ui=self.c_count },
    { hcenter=0, top=self.c_count, ui=c_avail },
    { left=5, top=c_avail, ui=vs1 },
    { left=0, top=vs1, ui=self.c_above },
    { hcenter=0, top=self.c_above, ui=c_chosen },
    { left=5, top=c_chosen, ui=vs2 },
    { left=0, top=vs2, ui=self.c_below },
    { left=0, bottom=0, ui=self.c_ok, hidden=true },
    { right=0, bottom=0, ui=self.c_cancel },
  }
  self:setFocus(self.c_above)
  self:setupUI(true, true)

  self.key:addBinds {
    EXIT = function()
      self:putBackIngredients()
      self.callback(nil)
      game:unregisterDialog(self)
    end,
  }
end

function _M:putBackIngredients()
  local inven = self.actor:getInven('INVEN')
  for _, o in ipairs(self.choices) do self.actor:addOrDropObject(inven, o) end
end

function _M:on_error()
  self:putBackIngredients()
end

function _M:selectItem(item, adding)
  if not item or not item.obj then return end

  -- If we've already pulled enough ingredients, don't pull anymore.
  if adding and self.choices.n >= self.needed then return end
  -- Shouldn't be possible, but just in case...
  if not adding and self.choices.n <= 0 then return end

  local inven = self.actor:getInven('INVEN')
  if adding then
    -- Pull ingredient from actor inventory into choices list.
    local o = self.actor:removeObject(inven, item.idx, 1)
    if o then
      self.choices.n = self.choices.n + 1
      self.choices[#self.choices+1] = o
      -- This will stack ingredient with others if needed.
      obj_util.sort_inven(self.choices)
    end
  else
    -- Put ingredient back into inventory from choices list.
    local o, last = item.obj:unstack()
    if last then table.remove(self.choices, item.idx) end
    self.actor:addOrDropObject(inven, o)
    self.actor:sortInven()
    self.choices.n = self.choices.n - 1
  end

  -- Rebuild both lists to reflect the changes.
  self:generateLists()
  self:updateCount()
  self:getUIElement(self.c_ok).hidden = (self.choices.n < self.needed)
end

function _M:finish()
  if self.choices.n ~= self.needed then return end
  self.callback(self.choices)
  game:unregisterDialog(self)
end

function _M:updateCount()
  local c = self.choices.n >= self.needed and '#LIGHT_GREEN#' or '#LIGHT_BLUE#'
  local fmt = '%s%d of %d selected#LAST#'
  local text = fmt:format(c, self.choices.n, self.needed)
  -- Slight Hack(TM):  Can't pass item=0 to :switchItem(), as it apparently
  -- uses cur_item=0 internally as "nothing selected".
  self.c_count:switchItem(self.choices.n+1, text)
end

function _M:generateLists()
  self.cands_list = { n=0 }
  self.choices_list = { n=0 }

  -- Build candidates list from actor inventory.
  self.actor:inventoryApply(self.actor:getInven('INVEN'), function(_, i, o)
    if util.alchemy.ok_ingredient(o, self.recipe, self.actor) then
      local num = o:getNumber()
      self.cands_list.n = self.cands_list.n + num
      self.cands_list[#self.cands_list+1] = {
	name = o:getName(),
	obj = o,
	num = num,
	idx = i,
      }
    end
  end)

  -- Build choices list from what we've chosen so far.
  for i, o in ipairs(self.choices) do
    local num = o:getNumber()
    self.choices_list.n = self.choices_list.n + num
    self.choices_list[#self.choices_list+1] = {
      name = o:getName(),
      obj = o,
      num = num,
      idx = i,
    }
  end

  if self.c_above then
    self.c_above.list = self.cands_list
    self.c_above:generate()
  end
  if self.c_below then
    self.c_below.list = self.choices_list
    self.c_below:generate()
  end
end
