-- $Id: AutomatizerClauseAction.lua 3638 2019-07-16 02:36:26Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local List = require 'engine.ui.List'
local AutomatizerMakeClause = require 'mod.dialogs.AutomatizerMakeClause'
local AutomatizerOneOfEditor = require 'mod.dialogs.AutomatizerOneOfEditor'

module(..., package.seeall, class.inherit(Dialog))

function _M:init(rule, clause, parent, child_idx, on_change_cb)
  self.rule = rule
  self.clause = clause
  self.parent = parent
  self.child_idx = child_idx
  self.on_change_cb = on_change_cb or function() end

  self:generateList()
  Dialog.init(self, 'Modify Condition', 1, 1)

  local list = List.new {
    width = self.max_w + 10,
    nb_items = #self.list,
    list = self.list,
    fct = function(item) self:use(item) end,
  }

  local function center_cb(w, h)
    local mx, my = core.mouse.get()
    self.force_x = mx - w/2
    self.force_y = my - (self.h - self.ih + list.fh/3)
  end

  self:loadUI { { left=0, top=0, ui=list } }
  self:setupUI(true, true, center_cb)

  self.key:addBinds {
    EXIT = function() game:unregisterDialog(self) end
  }
end

function _M:use(item)
  if not item or not item.action then return end
  self.key:triggerVirtual('EXIT')
  if item.action == 'modify' then
    local d = AutomatizerMakeClause.new(self.clause, function(cl)
      self.parent[self.child_idx] = cl
      self.on_change_cb()
    end)
    game:registerDialog(d)
  elseif item.action == 'modify_oneof_parent' then
    -- Vaguely Annoying Hack(TM):  Maybe we should consolidate all this
    -- information somewhere...
    local val_test = nil
    for _, test in ipairs(AutomatizerMakeClause.values) do
      if test.value == self.parent[1].clause then
	val_test = test
	break
      end
    end
    if not val_test then return end
    local vals = {}
    for i = 2, #self.parent do vals[i-1] = self.parent[i] end
    local function cb(ret)
      if ret then
	while #self.parent > 1 do table.remove(self.parent) end
	table.append(self.parent, ret)
	self.on_change_cb()
      end
    end
    game:registerDialog(AutomatizerOneOfEditor.new(val_test.name, val_test.group, val_test.grp_only, vals, cb))
  elseif item.action == 'absorb_single_child' then
    local child = self.clause[1]
    for k, _ in pairs(self.clause) do self.clause[k] = nil end
    table.merge(self.clause, child)
    self.on_change_cb()
  elseif item.action == 'wrap_clause' and item.wrap_test then
    local new_child = table.clone(self.clause)
    for k, _ in pairs(self.clause) do self.clause[k] = nil end
    self.clause.test = item.wrap_test
    self.clause[1] = new_child
    self.on_change_cb()
  elseif item.action == 'subclause' then
    local d = AutomatizerMakeClause.new(nil, function(cl)
      if #self.clause == 1 and self.clause[1].test == 'placeholder' then
	self.clause[1] = cl
      else
	table.insert(self.clause, cl)
      end
      self.on_change_cb()
    end)
    game:registerDialog(d)
  elseif item.action == 'remove' then
    if type(self.child_idx) == 'number' then
      if #self.parent == 1 and self.child_idx == 1 then
	self.parent[1] = { test='placeholder' }
      else
	table.remove(self.parent, self.child_idx)
      end
      self.on_change_cb()
    else
      -- ??? is this possible
      print('[DEBUG] AutomatizerClauseAction:use() non-numeric child idx', self.child_idx)
    end
  end
end

function _M:generateList()
  local list = {}
  self.max_w = 0

  if not self.clause then
    if self.parent and self.parent.test == 'oneof' then
      -- Slight Hack(TM):  Don't allow removing the only member of the
      -- one-of group.
      if #self.parent > 2 then
	list[#list+1] = { name='Remove this value', action='remove' }
      end
      list[#list+1] = { name='Modify this value list', action='modify_oneof_parent' }
    end
  elseif self.clause.test == 'placeholder' then
    list[#list+1] = { name='Replace placeholder with test', action='modify' }
  else
    if self.clause.test == 'and' or self.clause.test == 'or' then
      list[#list+1] = { name='Add new subtest', action='subclause' }
      if #self.clause == 1 then
	list[#list+1] = {
	  name = 'Remove "'..self.clause.test..'" around single subtest',
	  action = 'absorb_single_child',
	}
      end
    end
    list[#list+1] = { name='Modify or replace this test', action='modify' }
    if self.clause.test == 'and' or self.clause.test == 'or' then
      list[#list].name = ('Replace "%s" group with new test'):format(self.clause.test)
    end
    list[#list+1] = { name='Remove this test', action='remove' }
    local negate_action, negate_wrap = 'wrap_clause', 'not'
    if self.clause.test == 'not' then
      negate_action, negate_wrap = 'absorb_single_child', nil
    elseif self.parent and self.parent.test == 'not' then
      negate_action, negate_wrap = 'remove_parent_not', nil
    end
    list[#list+1] = { name='Negate this test', action=negate_action, wrap_test=negate_wrap }
    if self.clause.test ~= 'and' then
      list[#list+1] = { name='Wrap this test in an "and" group', action='wrap_clause', wrap_test='and' }
    end
    if self.clause.test ~= 'or' then
      list[#list+1] = { name='Wrap this test in an "or" group', action='wrap_clause', wrap_test='or' }
    end
  end

  for _, item in ipairs(list) do
    local w, _ = self.font:size(item.name)
    self.max_w = math.max(self.max_w, w)
  end

  self.list = list
end
