-- $Id: FontSettings.lua 4222 2021-06-02 02:23:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local Textzone = require 'engine.ui.Textzone'
local Button = require 'engine.ui.Button'
local List = require 'engine.ui.List'
local FontPackage = require 'engine.FontPackage'

module(..., package.seeall, class.inherit(Dialog))

_M.sizes = {
  { name=_t'Normal', size='normal' },
  { name=_t'Small', size='small' },
  { name=_t'Big', size='big' },
}

function _M:init(title, init_val)
  Dialog.init(self, title, 1, 1)

  local vals = init_val:split('/')
  self.sel_style, self.sel_size = vals[1], vals[2]

  local styles = FontPackage:list()

  self:setupDisplayTexts(styles, self.sel_style, 'id')
  self:setupDisplayTexts(self.sizes, self.sel_size, 'size')

  local c_style_head = Textzone.new {
    width = 250,
    auto_height = true,
    no_color_bleed = true,
    text = 'Font style';
  }
  local c_size_head = Textzone.new {
    width = 250,
    auto_height = true,
    no_color_bleed = true,
    text = 'Font size';
  }
  self.c_style = List.new {
    list = styles,
    width = 250,
    nb_items = 5,
    display_prop = 'display',
    scrollbar = true,
    fct = function(item)
      if item and item.id then
	self.sel_style = item.id
	self:setupDisplayTexts(self.c_style.list, self.sel_style, 'id', self.c_style)
      end
    end,
  }
  self.c_size = List.new {
    list = self.sizes,
    width = 250,
    nb_items = 5,
    display_prop = 'display',
    scrollbar = true,
    fct = function(item)
      if item and item.size then
	self.sel_size = item.size
	self:setupDisplayTexts(self.c_size.list, self.sel_size, 'size', self.c_size)
      end
    end,
  }

  --[[
  for i, style in ipairs(styles) do
    if style.id == init_style then self.c_style.sel = i end
  end
  for i, size in ipairs(self.sizes) do
    if size.size == init_size then self.c_size.sel = i end
  end
  ]]

  self.c_ok = Button.new {
    text = 'Apply',
    fct = function() self.key:triggerVirtual('ACCEPT') end,
  }
  self.c_cancel = Button.new {
    text = 'Cancel',
    fct = function() self.key:triggerVirtual('EXIT') end,
  }

  self:loadUI {
    { left=0, top=0, ui=c_style_head },
    { left=0, top=c_style_head, ui=self.c_style },
    { left=self.c_style, top=0, ui=c_size_head },
    { left=self.c_style, top=c_size_head, ui=self.c_size },
    { right=0, bottom=0, ui=self.c_cancel },
    { right=self.c_cancel, bottom=0, ui=self.c_ok },
  }
  self:setupUI(true, true)

  self.key:addBinds {
    ACCEPT = function()
      self:setupValues()
      game:unregisterDialog(self)
    end,
    EXIT = function() game:unregisterDialog(self) end,
  }
end

function _M:selectItem(c)
  self:setupDisplayTexts(c)
end

function _M:setupDisplayTexts(list, txt, fld, c)
  for i, item in ipairs(list) do
    if item and item.name and txt then
      local old_disp = item.display
      local sel = item[fld] == txt
      item.display = sel and '#{bold}#'..item.name..'#{normal}#' or item.name
      if c and old_disp and old_disp ~= item.display then c:drawItem(item) end
    end
  end
end

function _M:setupValues()
  --[[
  local style = self.c_style.list[self.c_style.sel].id
  local size = self.c_size.list[self.c_size.sel].size
  ]]
  self.value = self.sel_style .. '/' .. self.sel_size
end

function _M:get_value()
  return self.value
end

function _M:get_display_val()
  return self:displayValue(self.value)
end

function _M:displayValue(val)
  if type(val) ~= 'string' then return '???' end
  local vals = val:split('/')
  local style, size = vals[1], vals[2]

  for _, style_ in ipairs(FontPackage:list()) do
    if style_.id == style then style = style_.name end
  end
  for _, size_ in ipairs(self.sizes) do
    if size_.size == size then size = size_.name end
  end

  return style .. '/' .. size
end
