-- $Id: RawTextbox.lua 4290 2021-09-27 02:56:16Z dsb $
-- ToME - Tales of Maj'Eyal
-- Copyright © 2017-2021 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham ("Zizzo")
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Textbox = require 'engine.ui.Textbox'
local Entity = require 'engine.Entity'

-- A modified version of engine.ui.Textbox that avoids "live" display of
-- TString highlighing codes in the entered text by mocking up a suitable
-- "raw" rendering of the text.

module(..., package.seeall, class.inherit(Textbox))

function _M:init(t)
  -- Spectacularly Evil But Awesome Hack(TM):  Since apparently none of the
  -- text-displaying engine code can display a '#' character without
  -- thinking it's part of a TString highlighting sequence, we make an
  -- Entity with a '#' in it to display in its place, using deep wizardry
  -- to force the width of the entity to the width of the character.
  self.pound = Entity.new {
    display = '#',
    color = colors.WHITE,
    force_w = self.font_mono:size('#'),
    getEntityFinalSurface = function(e_self, tiles, w, h)
      return require('engine.Entity').getEntityFinalSurface(e_self, tiles, e_self.force_w, h)
    end,
  }
  self.pound_str = self.pound:getDisplayString()

  Textbox.init(self, t)
end

-- Overloaded to put highlighted text into our texture.
function _M:updateText()
  if not self.tmp[1] then self.tmp = {} end
  self.text = table.concat(self.tmp)

  local text
  local b, e = self.scroll, math.min(self.scroll + self.max_display - 1, #self.tmp)
  if not self.hide then 
    text = self:buildText(b, e)
  else
    text = string.rep('*', e - b + 1)
  end

  self.text_tex = self:drawFontLine(self.font_mono, text, self.fw)

  if self.on_change and self.old_text ~= self.text then
    self.on_change(self.text)
  end
  self.old_text = self.text
end

-- Builds and returns a subsequence of the entered text with our special
-- entity's display string inserted in all the right places.
function _M:buildText(b, e)
  local text = ''
  for i = b, e do
    text = text .. (self.tmp[i] == '#' and self.pound_str or self.tmp[i])
  end
  return text
end

-- Overloaded to use a subsequence of the highlighted text to compute the
-- cursor position.
function _M:display(x, y, nb_keyframes)
  local text_x, text_y = self.text_x, self.text_y
  self:displayTitle(x, y, nb_keyframes)

  if self.focused then
    self:drawFrame(self.frame_sel, x + self.title_w, y)
    local text = self:buildText(self.scroll, self.cursor - 1)
    local cursor_x = self.font_mono:size(text)
    self.texcursor.t:toScreenFull(x + self.text_x + cursor_x - (self.texcursor.w / 2) + 2, y + text_y + 5,
    self.texcursor.w, self.fh - 7, self.texcursor.tw, self.texcursor.th)
  else
    self:drawFrame(self.frame, x + self.title_w, y)
    if self.focus_decay then
      self:drawFrame(self.frame_sel, x + self.title_w, y, 1, 1, 1, self.focus_decay / self.focus_decay_max_d)
      self.focus_decay = self.focus_decay - nb_keyframes
      if self.focus_decay <= 0 then self.focus_decay = nil end
    end
  end
  if self.text_shadow then self:textureToScreen(self.text_tex, x+1 + self.text_x, y+1 + self.text_y, 0, 0, 0, self.text_shadow) end
  self:textureToScreen(self.text_tex, x+1 + self.text_x, y+1 + self.text_y)
end
