-- $Id: SubGroup.lua 3632 2019-07-14 01:32:38Z dsb $
-- TE4 - T-Engine 4
-- Copyright (C) 2009 - 2018 Nicolas Casalini
-- Copyright © 2018-2019 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Nicolas Casalini "DarkGod"
-- darkgod@te4.org

require "engine.class"
local KeyBind = require "engine.KeyBind"
local Base = require "engine.ui.Base"
local Focusable = require "engine.ui.Focusable"

-- A container of UI elements with layout mechanics similar to
-- engine.ui.Dialog, which can be loosely manipulated as a group.
-- Modified from engine.ui.SubDialog.
module(..., package.seeall, class.inherit(Base, Focusable))

function _M:init(w, h)
  self.alpha = self.alpha or 255
  self.color = self.color or {r=255, g=255, b=255}
  if skin then self.ui = skin end

  self.uis = {}
  self.ui_by_ui = {}
  self.focus_ui = nil
  self.focus_ui_id = 0

  self.first_display = true

  Base.init(self, {}, true)

  self:resize(w, h, true)
end

function _M:resize(w, h, nogen)
  self.w, self.h = math.floor(w), math.floor(h)
  self.ix, self.iy = 0, 0
  self.iw, self.ih = w, h

  if not nogen then self:generate() end
end

function _M:generate()
  local gamew, gameh = core.display.size()

  self.mouse:registerZone(0, 0, self.w, self.h, function(...) self:mouseEvent(...) end)
  self.key.receiveKey = function(_, ...) self:keyEvent(...) end
  self.key:addCommands{
    _TAB = function() self:moveFocus(1) end,
    _UP = function() self:moveFocus(-1) end,
    _DOWN = function() self:moveFocus(1) end,
    _LEFT = function() self:moveFocus(-1) end,
    _RIGHT = function() self:moveFocus(1) end,
  }
end

function _M:loadUI(t)
  if not t.no_reset then
    self.uis = {}
    self.ui_by_ui = {}
    self.focus_ui = nil
    self.focus_ui_id = 0
  end
  for i, ui in ipairs(t) do
    self.uis[#self.uis+1] = ui
    self.ui_by_ui[ui.ui] = ui

    if not self.focus_ui and ui.ui.can_focus then
      self:setSubFocus(i)
    elseif ui.ui.can_focus then
      ui.ui:setFocus(false)
    end
  end
end

function _M:setupUI(resizex, resizey, on_resize, addmw, addmh)
  local mw, mh = nil, nil
  local padding = 3 -- to not glue stuff to each other

  if resizex or resizey then
    mw, mh = 0, 0
    local addw, addh = 0, 0

    for i, ui in ipairs(self.uis) do
      if ui.top and type(ui.top) == "table" then
	ui.top = self.ui_by_ui[ui.top].top + self.ui_by_ui[ui.top].ui.h + padding
      end
      if ui.bottom and type(ui.bottom) == "table"
	then ui.bottom = self.ui_by_ui[ui.bottom].bottom + self.ui_by_ui[ui.bottom].ui.h + padding
	end
      if ui.left and type(ui.left) == "table" then
	ui.left = self.ui_by_ui[ui.left].left + self.ui_by_ui[ui.left].ui.w + padding
	end
      if ui.right and type(ui.right) == "table" then
	ui.right = self.ui_by_ui[ui.right].right + self.ui_by_ui[ui.right].ui.w + padding
      end
				
      if not ui.absolute then
	if ui.top then
	  mh = math.max(mh, ui.top + ui.ui.h + (ui.padding_h or 0))
	elseif ui.bottom then
	  addh = math.max(addh, ui.bottom + ui.ui.h + (ui.padding_h or 0))
	end

	if ui.left then
	  mw = math.max(mw, ui.left + ui.ui.w + (ui.padding_w or 0))
	elseif ui.right then
	  addw = math.max(addw, ui.right + ui.ui.w + (ui.padding_w or 0))
	end
      end
    end
    mw = mw + addw + 5 * 2 + (addmw or 0)

    local tw, th = 0, 0
    mw = math.max(tw + 6, mw)

    mh = mh + addh + 5 + 22 + 3 + (addmh or 0) + th

    local new_w = resizex and mw or self.w
    local new_h = resizey and mh or self.h
    if on_resize then on_resize(new_w, new_h) end
    self:resize(new_w, new_h)
  else
    if on_resize then on_resize(self.w, self.h) end
    self:resize(self.w, self.h)
  end

  for i, ui in ipairs(self.uis) do
    local ux, uy

    if not ui.absolute then
      ux, uy = self.ix, self.iy

      if ui.top then
	if type(ui.top) == "table" then ui.top = self.ui_by_ui[ui.top].y end
	uy = uy + ui.top
      elseif ui.bottom then
	if type(ui.bottom) == "table" then
	  ui.bottom = self.ui_by_ui[ui.bottom].y
	end
	uy = uy + self.ih - ui.bottom - ui.ui.h
      elseif ui.vcenter then
	if type(ui.vcenter) == "table" then
	  ui.vcenter = self.ui_by_ui[ui.vcenter].y + ui.vcenter.h
	end
	uy = uy + math.floor(self.ih / 2) + ui.vcenter - ui.ui.h / 2
      end

      if ui.left then
	if type(ui.left) == "table" then
	  ui.left = self.ui_by_ui[ui.left].x + ui.left.w
	end
	ux = ux + ui.left
      elseif ui.right then
	if type(ui.right) == "table" then
	  ui.right = self.ui_by_ui[ui.right].x
	end
	ux = ux + self.iw - ui.right - ui.ui.w
      elseif ui.hcenter then
	if type(ui.hcenter) == "table" then
	  ui.hcenter = self.ui_by_ui[ui.hcenter].x + ui.hcenter.w
	end
	ux = ux + math.floor(self.iw / 2) + ui.hcenter - ui.ui.w / 2
      end
    else
      ux, uy = 0, 0

      if ui.top then uy = uy + ui.top
      elseif ui.bottom then uy = uy + game.h - ui.bottom - ui.ui.h
      elseif ui.vcenter then uy = uy + math.floor(game.h / 2) + ui.vcenter - ui.ui.h / 2 end

      if ui.left then ux = ux + ui.left
      elseif ui.right then ux = ux + game.w - ui.right - ui.ui.w
      elseif ui.hcenter then ux = ux + math.floor(game.w / 2) + ui.hcenter - ui.ui.w / 2 end

      --ux = ux - self.display_x
      --uy = uy - self.display_y
    end

    ui.x = ux
    ui.y = uy
    ui.ui.mouse.delegate_offset_x = ux
    ui.ui.mouse.delegate_offset_y = uy
    -- Vaguely Annoying Hack(TM):  We don't have display_x/y yet; fake it.
    --ui.ui:positioned(ux, uy, self.display_x + ux, self.display_y + uy)
    ui.ui:positioned(ux, uy, ux, uy)
  end

  self.setuped = true
end

function _M:positioned(x, y, sx, sy)
  -- Position all our subcomponents relative to ourself.
  for _, ui in ipairs(self.uis) do
    ui.ui.mouse.delegate_offset_x = ui.x
    ui.ui.mouse.delegate_offset_y = ui.y
    ui.ui:positioned(ui.x, ui.y, sx + ui.x, sy + ui.y)
  end
end

function _M:replaceUI(oldui, newui)
  for i, ui in ipairs(self.uis) do
    if ui.ui == oldui then
      ui.ui = newui
      ui.ui.mouse.delegate_offset_x = ui.x
      ui.ui.mouse.delegate_offset_y = ui.y
      ui.ui:positioned(ui.x, ui.y, self.display_x + ui.x, self.display_y + ui.y)
    end
  end
end

function _M:setSubFocus(id)
  if self.focus_ui and self.focus_ui.ui.can_focus then
    self.focus_ui.ui:setFocus(false)
  end

  if type(id) == "table" then
    for i = 1, #self.uis do
      if self.uis[i].ui == id then id = i break end
    end
    if type(id) == "table" then return end
  end

  local ui = self.uis[id]
  if ui and not ui.ui.can_focus then return end
  self.focus_ui = ui
  self.focus_ui_id = id
  if ui then ui.ui:setFocus(true) end
end

function _M:moveUIElement(id, left, right, top, bottom)
  if type(id) == "table" then
    for i = 1, #self.uis do
      if self.uis[i].ui == id then id = i break end
    end
    if type(id) == "table" then return end
  end

  self.uis[id].left = left or self.uis[id].left
  self.uis[id].right = right or self.uis[id].right
  self.uis[id].top = top or self.uis[id].top
  self.uis[id].bottom = bottom or self.uis[id].bottom
end

function _M:getUIElement(id)
  if type(id) == "table" then
    for i = 1, #self.uis do
      if self.uis[i].ui == id then id = i break end
    end
    if type(id) == "table" then return end
  end

  return self.uis[id]
end

function _M:toggleDisplay(ui, show)
  if not self.ui_by_ui[ui] then return end
  self.ui_by_ui[ui].hidden = not show
end

function _M:moveFocus(v)
  local id = self.focus_ui_id
  local start = id or 1
  local cnt = 0
  id = util.boundWrap((id or 1) + v, 1, #self.uis)
  while start ~= id and cnt <= #self.uis do
    if self.uis[id] and self.uis[id].ui and self.uis[id].ui.can_focus and not self.uis[id].ui.no_keyboard_focus then
      self:setSubFocus(id)
      break
    end
    id = util.boundWrap(id + v, 1, #self.uis)
    cnt = cnt + 1
  end
end

function _M:on_focus(v)
  if not v then
    self.last_focus = self.focus_ui
    self:setSubFocus({})
  else
    self:setSubFocus(self.focus_ui)
  end
end

function _M:mouseEvent(button, x, y, xrel, yrel, bx, by, event)
  -- Look for focus
  for i = 1, #self.uis do
    local ui = self.uis[i]
    if ui.ui.can_focus and bx >= ui.x and bx <= ui.x + ui.ui.w and by >= ui.y and by <= ui.y + ui.ui.h then
      self:setSubFocus(i)

      -- Pass the event
      ui.ui.mouse:delegate(button, bx, by, xrel, yrel, bx, by, event)
      return
    end
  end
end

function _M:keyEvent(...)
  if not self.focus_ui or not self.focus_ui.ui.key:receiveKey(...) then
    KeyBind.receiveKey(self.key, ...)
  end
end

function _M:innerDisplayBack(x, y, nb_keyframes)
end
function _M:innerDisplay(x, y, nb_keyframes)
end

function _M:firstDisplay()
end

function _M:display(x, y, nb_keyframes)
  if self.__hidden then return end

  -- We translate opengl matrix to make the popup effect easily
  local ox, oy = x, y
  local hw, hh = math.floor(self.w / 2), math.floor(self.h / 2)
  local tx, ty = x + hw, y + hh
  x, y = -hw, -hh
  core.display.glTranslate(tx, ty, 0)

  self:innerDisplayBack(x, y, nb_keyframes, tx, ty)

  -- UI elements
  for i = 1, #self.uis do
    local ui = self.uis[i]
    if not ui.hidden then ui.ui:display(x + ui.x, y + ui.y, nb_keyframes, ox + ui.x, oy + ui.y) end
  end

  self:innerDisplay(x, y, nb_keyframes, tx, ty)

  if self.first_display then self:firstDisplay() self.first_display = false end

  -- Restiore normal opengl matrix
  core.display.glTranslate(-tx, -ty, 0)
end
